/* This file is part of cqual.
   Copyright (C) 2000-2003 The Regents of the University of California.

cqual is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

cqual is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with cqual; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */

#include <assert.h>
#include <string.h>
#include "hash.h"
#include "regions.h"
#include "utils.h"

struct bucket
{
  hash_key key;
  hash_data data;
  struct bucket *next;
};

#define scan_bucket(b, var) for (var = b; var; var = var->next)

#define num_valid_bytes(ht) (((ht)->size + 7) / 8)
#define is_valid(ht,hash) ((ht)->valid_bits[(hash) / 8] & (0x01 << ((hash) % 8)))

struct Hash_table
{
  region r;         		/* Region for this table */
  hash_fn hash;     		/* Function for hashing keys */
  keyeq_fn cmp;     		/* Function for comparing keys */

  unsigned long size;		/* Number of slots */
  unsigned long log2size;	/* log2 of size */
  unsigned long used;		/* Number of elements */
  char *valid_bits;             /* bit-vector of valid slots */
  hash_key *keys;		/* Array of (size) key slots */
  hash_data *vals;              /* Array of (size) val slots */
};

static void rehash(hash_table ht);

/* Make a new hash table, with size buckets initially.  Hash table
   elements are allocated in region rhash. */
hash_table make_hash_table(region r, bool needvals, unsigned long size, hash_fn hash,
			   keyeq_fn cmp)
{
  hash_table result;

  assert(size > 0);
  result = ralloc(r, struct Hash_table);
  result->r = r;
  result->hash = hash;
  result->cmp = cmp;
  result->size = 1;
  result->log2size = 0;
  /* Force size to a power of 2 */
  while (result->size < size)
    {
      result->size *= 2;
      result->log2size++;
    }
  result->used = 0;
  result->valid_bits = rarrayalloc(result->r, num_valid_bytes(result), char);
  result->keys = rarrayalloc(result->r, result->size, hash_key);
  if (needvals)
    result->vals = rarrayalloc(result->r, result->size, hash_data);

  return result;
}

/* Make a hash table for strings. */
hash_table make_string_hash_table(region rhash, unsigned long size)
{
  return make_hash_table(rhash, TRUE, size, (hash_fn) string_hash,
			 (keyeq_fn) string_eq);
}

/* Zero out ht.  Doesn't reclaim bucket space. */
void hash_table_reset(hash_table ht)
{
  unsigned long i;

  ht->used = 0;
  for (i = 0; i < num_valid_bytes(ht); i++)
    ht->valid_bits[i] = 0;
}

/* Return the number of entries in ht */
unsigned long hash_table_size(hash_table ht)
{
  return ht->used;
}

#define MAGIC 2*0.6180339987

#define LLMAGIC ((unsigned long long)(MAGIC * (1ULL << (8 * sizeof(unsigned long) - 1))))

/* Return the bucket corresponding to k in ht */
static inline unsigned long find_bucket(hash_table ht, hash_key k)
{
  unsigned long hash;

  hash = ht->hash(k);
  hash = hash * LLMAGIC;
  hash = hash >> (8 * sizeof(unsigned long) - ht->log2size);
  if (ht->size == 1)
    hash = 0;
  /* hash = hash % ht->size; */
  assert(hash < ht->size);
  return hash;
}

/* Given a comparison function which agrees with our hash_function,
   search for the given element. */
bool hash_table_hash_search(hash_table ht, keyeq_fn cmp, 
			    hash_key k, hash_data *d)
{
  unsigned long hash;

  assert(ht->vals);

  hash = find_bucket(ht, k);
  while (is_valid(ht, hash))
    {
      if (cmp(k, ht->keys[hash]))
	{
	  if (d)
	    *d = ht->vals[hash];
	  return TRUE;
	}
      hash = (hash + 1) % ht->size;
    }
  return FALSE;
}

/* Given a comparison function which agrees with our hash_function,
   search for the given element. */
bool hash_table_hash_search_keys(hash_table ht, keyeq_fn cmp, 
			    hash_key k, hash_key *d)
{
  unsigned long hash;

  hash = find_bucket(ht, k);
  while (is_valid(ht, hash))
    {
      if (cmp(k, ht->keys[hash]))
	{
	  if (d)
	    *d = ht->keys[hash];
	  return TRUE;
	}
      hash = (hash + 1) % ht->size;
    }
  return FALSE;
}

/* Lookup k in ht.  Returns corresponding data in *d, and function
   result is TRUE if the k was in ht, false otherwise. */
bool hash_table_lookup(hash_table ht, hash_key k, hash_data *d)
{
  return hash_table_hash_search(ht, ht->cmp, k, d);
}

/* Add k:d to ht.  If k was already in ht, replace old entry by k:d.
   Rehash if necessary.  Returns TRUE if k was not already in ht. */
bool hash_table_insert(hash_table ht, hash_key k, hash_data d)
{
  unsigned long hash;

  if (ht->used >= 3 * ht->size / 4)
    rehash(ht);

  hash = find_bucket(ht, k);
  while (is_valid(ht, hash))
    {
      if (ht->cmp(k, ht->keys[hash]))
	{
	  if (ht->vals)
	    ht->vals[hash] = d;
	  return FALSE; /* Replace */
	}
      hash = (hash + 1) % ht->size;
    }
  ht->keys[hash] = k;
  if (ht->vals)
    ht->vals[hash] = d;
  ht->valid_bits[hash / 8] |= (0x01 << (hash % 8));
  ht->used++;
  return TRUE; /* New key */
}

bool is_between(unsigned long lo, unsigned long mid, unsigned long hi)
{
  if (lo <= hi)
    return lo <= mid && mid < hi;
  return lo <= mid || mid < hi;
}

/* Remove mapping for k in ht.  Returns TRUE if k was in ht. */
bool hash_table_remove(hash_table ht, hash_key k) 
{
  unsigned long hash;
  unsigned long gap;

  hash = find_bucket(ht, k);
  while (is_valid(ht, hash))
    {
      if (ht->cmp(k, ht->keys[hash]))
	{
	  ht->valid_bits[hash / 8] ^= (0x01 << (hash % 8));
	  ht->used--;

	  /* Mind the gap */
	  gap = hash;
	  for (hash = (hash + 1) % ht->size; is_valid(ht, hash); hash = (hash + 1) % ht->size)
	    if (is_between(find_bucket(ht, ht->keys[hash]), gap, hash))
	      {
		ht->keys[gap] = ht->keys[hash];
		if (ht->vals)
		  ht->vals[gap] = ht->vals[hash];
		ht->valid_bits[gap/8] |= (0x01 << (gap % 8));
		ht->valid_bits[hash/8] ^= (0x01 << (hash % 8));
		gap = hash;
	      }

	  return TRUE;
	}
      hash = (hash + 1) % ht->size;
    }
  return FALSE;
}

/* Return a copy of ht */
hash_table hash_table_copy(region r, hash_table ht)
{
  hash_table result;

  result = make_hash_table(r, ht->vals != NULL, ht->size, ht->hash, ht->cmp);
  result->used = ht->used;

  memcpy(result->valid_bits, ht->valid_bits, num_valid_bytes(ht));
  memcpy(result->keys, ht->keys, ht->size * sizeof(ht->keys[0]));
  if (ht->vals)
    memcpy(result->vals, ht->vals, ht->size * sizeof(ht->vals[0]));

  return result;
}

/* Increase size of ht (double it) and reinsert all the elements */
static void rehash(hash_table ht)
{
  hash_key * old_keys;
  hash_data * old_vals;
  char * old_valids;
  unsigned long old_table_size;
  unsigned long i;

#ifdef DEBUG
  printf("Rehash table size=%ld, used=%ld\n", ht->size, ht->used);
#endif

  old_table_size = ht->size;
  old_keys = ht->keys;
  old_vals = ht->vals;
  old_valids = ht->valid_bits;

  ht->size = ht->size*2;
  ht->log2size = ht->log2size + 1;
  ht->used = 0;
  ht->keys = rarrayalloc(ht->r, ht->size, hash_key);
  if (ht->vals)
    ht->vals = rarrayalloc(ht->r, ht->size, hash_data);
  ht->valid_bits = rarrayalloc(ht->r, num_valid_bytes(ht), char);

  for (i = 0; i < old_table_size; i++)
    if (old_valids[i/8] & (0x01 << (i % 8)))
      hash_table_insert(ht, old_keys[i], old_vals ? old_vals[i] : NULL);
}

/* Begin scanning ht */
void hash_table_scan(hash_table ht, hash_table_scanner *hts)
{
  hts->ht = ht;
  hts->i = 0;
}

/* Get next elt in table, storing the elt in *k and *d if k and d are
   non-NULL, respectively.  Returns TRUE if there is a next elt, FALSE
   otherwise. */
bool hash_table_next(hash_table_scanner *hts, hash_key *k, hash_data *d)
{
  while (hts->i < hts->ht->size && !is_valid(hts->ht, hts->i))
    hts->i++;

  if (hts->i < hts->ht->size)
    {
      if (k)
	*k = hts->ht->keys[hts->i];
      if (d)
	*d = hts->ht->vals ? hts->ht->vals[hts->i] : NULL;
      hts->i++;
      return TRUE;
    }

  return FALSE;
}

/* Apply f to all elements of ht, in some arbitrary order */
void hash_table_apply(hash_table ht, hash_apply_fn f, void *arg)
{
  unsigned long i;

  for (i = 0; i < ht->size; i++)
    if (is_valid(ht, i))
      f(ht->keys[i], ht->vals ? ht->vals[i] : NULL, arg);
}

/* Map f to all elements on ht, creating a new hash table in region r */
hash_table hash_table_map(region r, hash_table ht, hash_map_fn f, void *arg)
{
  unsigned long i;
  hash_table result;

  assert(ht->vals);

  result = hash_table_copy(r, ht);
  
  for (i = 0; i < result->size; i++)
    if (is_valid(result, i))
      result->vals[i] = f(result->keys[i], result->vals[i], arg);

  return result;
}

/* Re-arrange hash-table in place after some elements may have their 
   hash value change. */
void hash_table_update_hashes(hash_table ht)
{
  unsigned long i;
  hash_key *tkeys;
  hash_data *tvals = NULL;
  char *tvalids;

  tkeys = malloc(ht->size * sizeof(ht->keys[0]));
  if (ht->vals)
    tvals = malloc(ht->size * sizeof(ht->vals[0]));
  tvalids = malloc(num_valid_bytes(ht));
  memcpy(tkeys, ht->keys, ht->size * sizeof(ht->keys[0]));
  if (ht->vals)
    memcpy(tvals, ht->vals, ht->size * sizeof(ht->vals[0]));
  memcpy(tvalids, ht->valid_bits, num_valid_bytes(ht));

  memset(ht->valid_bits, 0, num_valid_bytes(ht));
  ht->used = 0;

  for (i = 0; i < ht->size; i++)
    if (tvalids[i / 8] & (0x01 << (i % 8)))
      hash_table_insert(ht, tkeys[i], tvals ? tvals[i] : NULL);

  free(tkeys);
  if (tvals)
    free(tvals);
  free(tvalids);
}

static keycmp_fn cur_cmp = NULL;

int entry_cmp(const void *a, const void *b)
{
  struct sorted_entry *ae = (struct sorted_entry *) a;
  struct sorted_entry *be = (struct sorted_entry *) b;
  return cur_cmp(ae->k, be->k);
}

/* Begin scanning ht in sorted order according to f */
void hash_table_scan_sorted(hash_table ht, keycmp_fn f,
			    hash_table_scanner_sorted *htss)
{
  hash_table_scanner hts;
  unsigned long i;

  htss->r = newregion();
  htss->size = hash_table_size(ht);
  htss->entries = rarrayalloc(htss->r, htss->size, struct sorted_entry);
  htss->i = 0;

  hash_table_scan(ht, &hts);
  i = 0;
  while (hash_table_next(&hts, &htss->entries[i].k,
			 &htss->entries[i].d))
    i++;
  assert(i == htss->size);
  cur_cmp = f;
  qsort(htss->entries, htss->size, sizeof(struct sorted_entry), entry_cmp);
  cur_cmp = NULL;
}

/* Just like hash_table_next, but scans in sorted order */
bool hash_table_next_sorted(hash_table_scanner_sorted *htss, hash_key *k,
			    hash_data *d)
{
  if (htss->i < htss->size)
    {
      if (k)
	*k = htss->entries[htss->i].k;
      if (d)
	*d = htss->entries[htss->i].d;
      htss->i++;
      return TRUE;
    }
  else
    {
      deleteregion(htss->r);
      htss->r = NULL;
      return FALSE;
    }
}
