/* This file is part of cqual.
   Copyright (C) 2004 University of Maryland

   cqual is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   cqual is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with cqual; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA. */

#define DEBUG printf

#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include "libqual_Native.h"
#include "libqual.h"
#include "quals.h"
#include "location.h"
#include "hash.h"
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_MALLOC_MALLOC_H
#include <malloc/malloc.h>
#endif

/* Name of program invoked, sans directories.  */
char *progname = "libqual"; 

/* XXX:  Shouldn't need these, but... */
typedef enum {sev_err, sev_warn, sev_info} severity;

/* size of biggest enum for jqual ffi anaysis - declared in quals.c */
extern int maxenum; 

extern void profile(void);

static hash_table interned_strings = NULL;
static region libqual_region = NULL;

#define MAX_STR 256 /* Max # of *bytes* passed as strings.
		       No allowance made for unicode! */

static int max_err = 0;
static const char **qual_msgs = NULL;

/*************************************************************************/
/*                                                                       */
/* Utilities                                                             */
/*                                                                       */
/*************************************************************************/

void debug_print_loc(location l){
  fprintf(stderr,"%s: %d",l->filename,l->lineno);
}

static int print_to_err(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  return fprintf(stderr, fmt, args);
}

/* Convert a jstring into a C string */
static const char *jstring_to_charp(JNIEnv *env, jstring s) {
  return (*env)->GetStringUTFChars(env, s, 0);
  /* It's actually slower to use the code below, so avoid it...
  char buf[MAX_STR];
  jsize len = (*env)->GetStringLength(env, s);
  if (len < MAX_STR)
    (*env)->GetStringUTFRegion(env, s, 0, len, buf);
  else
    fail("Increase buffer size in jstring_to_charp");
  return rstrdup(libqual_region, buf);
  */
}

/*************************************************************************/
/*                                                                       */
/* Library Initialization/Completion                                     */
/*                                                                       */
/*************************************************************************/

JNIEXPORT void JNICALL Java_cqual_Native_initCQual
(JNIEnv *env, jclass obj, jstring j_lattice, jstring j_ffiquals,
 jboolean j_ffiSubtypes, jboolean j_useEnums, jint j_startptr, jint j_maxptrs)
{
  const char *lattice = NULL;
  const char *ffiquals = NULL;
  region_init();
  init_qerror();
  init_quals();
  libqual_region = newregion();
  interned_strings = make_string_hash_table(libqual_region, 17);
 
  lattice = jstring_to_charp(env, j_lattice);
  if (lattice == NULL) return;

  if(strlen(lattice)>0) {
    load_config_file_quals(lattice);
  } else {
    fprintf(stderr,"No lattice path was specified. Use jqual preferences to set the lattice path.\n");
  }


  if (j_ffiquals != NULL) {
    ffiquals = jstring_to_charp(env, j_ffiquals);
    if(strlen(ffiquals)>0) {
      fprintf(stderr,"FFI Quals file: %s\n", ffiquals);
      fprintf(stderr,"FFI ptr range: %d..%d\n", (int)j_startptr, (int)j_maxptrs);
      fprintf(stderr,"FFI use enums?: %s\n", ((bool)j_useEnums)?"true":"false");

      load_opaque_quals(ffiquals, (bool)j_ffiSubtypes, (bool)j_useEnums, (int)j_startptr, (int)j_maxptrs);
    } else {
      fprintf(stderr,"No FFI Quals file: %s\n", ffiquals);
    }
  }

  end_define_pos(); /* Allow cqual to run with no qualifiers */
  //init_analyze(); // only for analyze.c
}

JNIEXPORT jstring JNICALL Java_cqual_Native_finishCQual
(JNIEnv *env, jobject obj)
{
  int orig_stdout_fd = dup(fileno(stdout));
  int orig_stderr_fd = dup(fileno(stderr));
	
  char *prefix = "/tmp/";
  char fileName[512] = "/tmp/jqual.tmp";
  int cnt;
  int outToFile = FALSE;
  struct stat s;


  if ( outToFile){
    for(cnt = 0; (cnt < 999999) && (stat(fileName, &s) == 0); ++cnt)
      sprintf(fileName, "%sjqual%d.tmp", prefix, cnt);

    if(stat(fileName, &s) == -1) {

      fflush(stdout);
      fflush(stderr);
      FILE *outFile = fopen(fileName, "w");
      if(outFile) dup2(fileno(outFile), fileno(stdout));
      if(outFile) dup2(fileno(outFile), fileno(stderr));
      if(outFile) {
	finish_quals();
	fflush(outFile);
	fclose(outFile);			
	// restore orignal stderr, stdout
	dup2(orig_stdout_fd, fileno(stdout));
	dup2(orig_stderr_fd, fileno(stderr));
	fprintf(stderr,"CQUAL-OUTPUT-DONE: %s\n", fileName);
	
	return (*env)->NewStringUTF(env, fileName);
      } else {
	fprintf(stderr,"libqual: error creating jqual output file\n");
      } 
    } else {
      fprintf(stderr,"libqual: error creating jqual output file - file exists?\n");
    }
  } else {
    finish_quals();
  }
  return NULL;
}

/*************************************************************************/
/*                                                                       */
/* Library Utilities                                                     */
/*                                                                       */
/*************************************************************************/

JNIEXPORT void JNICALL Java_cqual_Native_resetLibQual
(JNIEnv* env, jclass cls)
{
  reset_libqual();
  if (libqual_region)
    deleteregion(libqual_region);
  libqual_region = NULL;
  interned_strings = NULL;
  max_err = 0;
  qual_msgs = NULL;
}

JNIEXPORT void JNICALL Java_cqual_Native_setPolyFlag
(JNIEnv *env, jobject obj, jboolean b)
{
  flag_poly = (bool)b;
}

JNIEXPORT jint JNICALL Java_cqual_Native_getMaxEnum
(JNIEnv *env, jclass obj) 
{
  return (jint)maxenum;

}

JNIEXPORT void JNICALL Java_cqual_Native_register_1messages
  (JNIEnv *env, jclass cls, jobjectArray arr)
{
  int len = (*env)->GetArrayLength(env, arr);
  int i;
  max_err = len;
  qual_msgs = rarrayalloc(libqual_region, len, char *);
  for (i = 0; i < len; i++) {
    jstring s = (*env)->GetObjectArrayElement(env, arr, i);
    qual_msgs[i] = jstring_to_charp(env, s);
  }
}


JNIEXPORT jlong JNICALL Java_cqual_Native_getpid
(JNIEnv* env, jclass cls)
{
  return getpid();
}

JNIEXPORT void JNICALL Java_cqual_Native_profile
(JNIEnv *env, jclass cls)
{
  printf("Malloc info\n");
#ifdef HAVE_MALLOC_ZONE_STATISTICS
  malloc_statistics_t zstats;
  struct mstats stats = mstats();
  printf("  bytes used  %luK\n"
	 "  bytes total %luK\n"
	 "  bytes free  %luK\n",
	 stats.bytes_total/1024,
	 stats.bytes_used/1024,
	 stats.bytes_free/1024);
  malloc_zone_statistics(NULL, &zstats);
  printf("  size_in_use     %luK\n"
         "  max_size_in_use %luK\n"
         "  size_allocated  %luK\n",
	 zstats.size_in_use/1024, zstats.max_size_in_use/1024, zstats.size_allocated/1024);
#elif defined HAVE_MALLINFO
  struct mallinfo info = mallinfo();
  printf("  total bytes       %luK\n"
	 "    mmap'd bytes    %luK\n",
	 info.uordblks/1024, info.hblkhd/1024);
#endif

  fflush(stdout);
  /*  profile(); */
}

/*************************************************************************/
/*                                                                       */
/* Qualifiers                                                            */
/*                                                                       */
/*************************************************************************/

JNIEXPORT jint JNICALL Java_cqual_Native_makeQvar
(JNIEnv *env, jobject obj, jstring jname, jint location_index,
 jboolean preferred, jboolean global)
{
  qual new_qual;
  const char *str = jstring_to_charp(env,  jname);
  if (str == NULL) return;

  new_qual = make_qvar(str, (location)location_index, (bool)preferred, (bool)global);
  return (int)new_qual;
}

JNIEXPORT jboolean JNICALL Java_cqual_Native_variableQual
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  return variable_qual((qual)nativeIndex);
}

JNIEXPORT jboolean JNICALL Java_cqual_Native_constantQual
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  return constant_qual((qual)nativeIndex);
}

JNIEXPORT jint JNICALL Java_cqual_Native_findQual
(JNIEnv *env, jobject obj, jstring jstr)
{
  qual new_qual;
  const char *str = (*env)->GetStringUTFChars(env, jstr, 0);
  if (str == NULL) return;

  new_qual = find_qual(str);
 (*env)->ReleaseStringUTFChars(env, jstr, str);
  return (int)new_qual;
}

JNIEXPORT jint JNICALL Java_cqual_Native_signQual
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  switch (sign_qual((qual) nativeIndex)) {
    case sign_pos: return 1;
    case sign_neg: return -1;
    case sign_eq: return 0;
    default: fail("Unknown value from sign_qual");
  }
}

JNIEXPORT jint JNICALL Java_cqual_Native_levelQual
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  switch (level_qual((qual) nativeIndex)) {
    case level_value: return 0;
    case level_ref: return 1;
    default: fail("Unknown value from sign_qual");
  }
}


JNIEXPORT jstring JNICALL Java_cqual_Native_nameQual
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  const char *s = name_qual((qual) nativeIndex);
  return (*env)->NewStringUTF(env, s);
}

JNIEXPORT void JNICALL Java_cqual_Native_setQualName
(JNIEnv *env, jobject obj,
  jint nativeIndex, jstring jname)
{
  const char *s = jstring_to_charp(env, jname);
  if (s == NULL) return;
  set_qual_name((qual) nativeIndex, s);
}

JNIEXPORT void JNICALL Java_cqual_Native_getQualifierPath
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  print_qual_paths((qual)nativeIndex);
}
		
/*

JNIEXPORT void JNICALL Java_cqual_Native_getQualifierPath
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  print_qual(&print_to_err, (qual)nativeIndex , &open_qgate);
}
*/

JNIEXPORT void JNICALL Java_cqual_Native_makeQvarGlobal
(JNIEnv *env, jobject obj, jint q_in, jboolean global)
{
  qual q = (qual)q_in;
  set_global_qual (q, (bool)global);
}

JNIEXPORT void JNICALL Java_cqual_Native_makeQvarPreferred
(JNIEnv *env, jobject obj, jint q_in, jboolean preferred)
{
  qual q = (qual)q_in;
  set_preferred_qual (q, (bool)preferred);
}

/*************************************************************************/
/*                                                                       */
/* Constraints                                                           */
/*                                                                       */
/*************************************************************************/


JNIEXPORT void JNICALL Java_cqual_Native_mkeqQual
(JNIEnv *env, jobject obj, jint loc, jint left, jint right, jint err)
{
  assert (err >= 0 && err <= max_err);
  mkeq_qual((location) loc, &open_qgate, (qual) left, (qual) right,
	    qual_msgs[err]);
}


JNIEXPORT void JNICALL Java_cqual_Native_mkleqQual
(JNIEnv *env, jobject obj, jint loc, jint left, jint right, jint err)
{
  assert (err >= 0 && err <= max_err);
  mkleq_qual((location) loc, &open_qgate, (qual) left, (qual) right,
	     qual_msgs[err]);
}

JNIEXPORT void JNICALL Java_cqual_Native_mkeqQualGated
(JNIEnv *env, jobject obj, jint qgate, jint loc, jint left, jint right,
 jint err)
{
  assert (err >= 0 && err <= max_err);
  mkeq_qual((location) loc, (qual_gate*) qgate, (qual) left, (qual) right,
	     qual_msgs[err]);
}

JNIEXPORT void JNICALL Java_cqual_Native_mkleqQualGated
(JNIEnv *env, jobject obj, jint qgate, jint loc, jint left, jint right,
 jint err)
{
  assert (err >= 0 && err <= max_err);
  mkleq_qual((location) loc, (qual_gate*) qgate, (qual) left, (qual) right,
	     qual_msgs[err]);
}

JNIEXPORT void JNICALL Java_cqual_Native_mkleqInst
(JNIEnv *env, jobject obj, jint qgate, jint loc, jint p, jint left, jint right,
 jint err)
{
  assert (err >= 0 && err <= max_err);

  mkinst_qual((location) loc, (qual_gate*) qgate, (qual) left, (qual) right,
	      (polarity) p, qual_msgs[err]);
}


/*************************************************************************/
/*                                                                       */
/* Locations                                                             */
/*                                                                       */
/*************************************************************************/

JNIEXPORT jint JNICALL Java_cqual_Native_allocLocation
(JNIEnv *env, jobject obj)
{
  static int id_counter = 0;

  location new_loc = malloc(sizeof(struct Location));
  new_loc->lineno = 0;
  new_loc->filepos = 0;
  new_loc->in_system_header = 0; // always false for now
  new_loc->location_index = ++id_counter;
  return (int)new_loc;
}

JNIEXPORT jint JNICALL Java_cqual_Native_allocQGate
(JNIEnv *env, jobject obj, jboolean forward, jboolean backward)
{
  static int id_counter = 0;
  qual_gate *qgate = (qual_gate*)malloc(sizeof (qual_gate));
  qgate->backward = backward;
  qgate->forward = forward;
  memset (qgate->mask, 0, sizeof (qgate->mask));
  return (int)qgate;
}

JNIEXPORT void JNICALL Java_cqual_Native_addQualToGate
(JNIEnv *env, jobject obj, jint qgate, jint q, jboolean allow)
{
  qual_gate_set_qual ((qual_gate *)qgate, (qual) q, (bool) allow);
}

JNIEXPORT void JNICALL Java_cqual_Native_qualGateSetAll
(JNIEnv *env, jobject obj, jint qgate, jboolean allow){
  gate_word newval = ((int)allow) ? -1 : 0;
  int i;

  for (i = 0; i < GATE_WORDS; i++)
    ((qual_gate *)qgate)->mask[i] = newval & open_qgate.mask[i];
}

JNIEXPORT jlong JNICALL Java_cqual_Native_getLineFromLocation
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  return ((location)nativeIndex)->lineno;
}


JNIEXPORT jlong JNICALL Java_cqual_Native_getFilePositionFromLocation
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  return ((location)nativeIndex)->filepos;
}

JNIEXPORT jstring JNICALL Java_cqual_Native_getFileNameFromLocation
(JNIEnv *env, jobject obj, jint nativeIndex)
{
  if(((location)nativeIndex)->filename){
    //    fprintf(stderr,"HERE2 - ");
    //    if (((location)nativeIndex)->filename == NULL)
      //      fprintf(stderr,"filename is null\n");
    //    else
    //      fprintf(stderr,"filename is %s\n", ((location)nativeIndex)->filename);
    jstring jstr =  (*env)->NewStringUTF(env, ((location)nativeIndex)->filename);
    //    fprintf(stderr,"HERE3\n");
    return jstr;
  }
  //  fprintf(stderr,"HERE4\n");
  return (*env)->NewStringUTF(env, "(no file)");
}

JNIEXPORT void JNICALL Java_cqual_Native_setLineForLocation
(JNIEnv *env, jobject obj, jint nativeIndex, jlong line)
{
  ((location)nativeIndex)->lineno = line;
}

JNIEXPORT void JNICALL Java_cqual_Native_setFilePositionForLocation
(JNIEnv *env, jobject obj, jint nativeIndex, jlong pos)
{
  ((location)nativeIndex)->filepos = pos;
}

JNIEXPORT void JNICALL Java_cqual_Native_setFileNameForLocation
(JNIEnv *env, jobject obj, jint nativeIndex, jstring jfilename)
{
  // is this right?
  //if(((location)nativeIndex)->filename)
  //free(((location)nativeIndex)->filename); // this wasn't malloc'ed
  //  const char *utf_j = (*env)->GetStringUTFChars(env, jfilename, 0);
  const char *utf_j = jstring_to_charp(env, jfilename);
  if (utf_j == NULL) return;

  ((location)nativeIndex)->filename = (char *) utf_j;
}

/*************************************************************************/
/*                                                                       */
/* Analysis Results                                                      */
/*                                                                       */
/*************************************************************************/

struct qList{
  int count;
  qual q;
  struct qList* n;
};

void storeQualifierIndex(qual q, void* arg){
  struct qList** b = ( struct qList**)arg;
  struct qList* bounds = *b;
  struct qList* new = ( struct qList*)malloc(sizeof(struct qList));
  if (q == NULL){
    fprintf(stderr,"Got a null q\n");
    return;
  }
  if (bounds == NULL)
    new->count = 1;
  else
    new->count = bounds->count+1;
  new->q = q;
  new->n = bounds;
  *b = new;
}

struct jQualifierInfo{
  JNIEnv *env;
  jobject q;
};

void addInferredQualifier(qual q, void* arg){
  struct jQualifierInfo* target = (struct jQualifierInfo*) arg;
  JNIEnv * env = target->env;
  jobject native_q = target->q;

  jclass c = (*env)->GetObjectClass(env, native_q);
  /*
  jfieldID fid = (*env)->GetFieldID(env, c, "nativeIndex", "I");
   if (fid == NULL){
          fprintf(stderr,"Qualifier not found\n");
	  return;
   }
  jint i = (*env)->GetIntField(env, q, fid);
  */

  jmethodID mid = (*env)->GetMethodID(env, c, "callbackAddInferredQualifier", "(ILjava/lang/String;)V");
  if (mid == NULL) {
          fprintf(stderr,"method not found\n");
  } else {
    (*env)->CallVoidMethod(env, native_q, mid, (jint)q, (*env)->NewStringUTF(env, name_qual(q)));
  }
}

void printit(qual q, void* arg){
  printf("inferred %d\n",(int)q);
}


JNIEXPORT void JNICALL Java_cqual_Native_getInferredQualifiers
(JNIEnv * env, jclass cls, jobject native_q)
{
  struct jQualifierInfo target;
  target.env = env;
  target.q = native_q;

  
  jclass c = (*env)->GetObjectClass(env, native_q);
  jfieldID fid = (*env)->GetFieldID(env, c, "nativeIndex", "I");
  if (fid == NULL){
          fprintf(stderr,"Qualifier not found\n");
	  return;
   }
   qual q = (qual) (*env)->GetIntField(env, native_q, fid);
  if (constant_qual(q)){
    printf("Constant\n");
  } else {
    qual_gate* qgate =  (qual_gate*) malloc(sizeof(qual_gate));
    struct qList *bounds = NULL;

    scan_qual_bounds( q, &fi_qgate,(qual_traverse_fn)addInferredQualifier , (void*)&target);
    //scan_qual_bounds( q, &fi_qgate,(qual_traverse_fn)printit , (void*)&target);

  }

  return;
}

JNIEXPORT jboolean JNICALL Java_cqual_Native_leqQual
(JNIEnv *env, jclass obj, jint nativeLeftQualIndex, jint nativeRightQualIndex) {
  return leq_qual((qual) nativeLeftQualIndex, (qual) nativeRightQualIndex);
}

JNIEXPORT jboolean JNICALL Java_cqual_Native_mayLeqQual
(JNIEnv *env, jclass obj, jint nativeLeftQualIndex, jint nativeRightQualIndex) {
  return may_leq_qual((qual) nativeLeftQualIndex, (qual) nativeRightQualIndex);
}

JNIEXPORT jboolean JNICALL Java_cqual_Native_csLeqQual
(JNIEnv *env, jclass obj, jint nativeLeftQualIndex, jint nativeRightQualIndex) {
  return cs_leq_qual((qual) nativeLeftQualIndex, (qual) nativeRightQualIndex);
}

JNIEXPORT jboolean JNICALL Java_cqual_Native_csMayLeqQual
(JNIEnv *env, jclass obj, jint nativeLeftQualIndex, jint nativeRightQualIndex) {
  return cs_may_leq_qual((qual) nativeLeftQualIndex, (qual) nativeRightQualIndex);
}


void cqual_test() {
  //init quals with test lattice
  const char *lattice = "test_lattice";

  region_init();
  init_qerror();
  init_quals();
  libqual_region = newregion();
  interned_strings = make_string_hash_table(libqual_region, 17);
 
  load_config_file_quals(lattice);
  end_define_pos();

  flag_poly = TRUE;

  //should be ok to share one location
  location loc = malloc(sizeof(struct Location));
  loc->location_index = 0;

  /*

  //make qual vars
qual cConsRet = make_qvar("C_cons_ret", loc, TRUE, FALSE);
qual OBJECT = make_qvar("JAVA.LANG.OBJECT", loc, TRUE, FALSE);
qual METHODSIG = make_qvar("JAVARITESTSUITE.METHODSIG", loc, TRUE, FALSE);
qual METHODSIG_C = make_qvar("JAVARITESTSUITE.METHODSIG.C", loc, TRUE, FALSE);
qual METHODSIG_C_x = make_qvar("JAVARITESTSUITE.METHODSIG.C.x", loc, TRUE, FALSE);
qual METHODSIG_C_x_ref = make_qvar("JAVARITESTSUITE.METHODSIG.C.x_ref", loc, TRUE, FALSE);
qual MethodSig_f0_invRet_26 = make_qvar("LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_invRet@26", loc, TRUE, FALSE);
qual MethodSig_f0_ret = make_qvar("LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_ret", loc, TRUE, FALSE);
qual MethodSig_f0_this = make_qvar("LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_this", loc, TRUE, FALSE);
qual MethodSig_f1_this = make_qvar("LjavariTestSuite/MethodSig;.f1()V_this", loc, TRUE, FALSE);
qual MethodSig_cons_ret = make_qvar("MethodSig_cons_ret", loc, TRUE, FALSE);
qual Object_cons_ret = make_qvar("Object_cons_ret", loc, TRUE, FALSE);

  //find qual constants
  qual mutable = find_qual("mutable");
  if (mutable == NULL) 
    fail("could not find mutable");

  qual readonly = find_qual("readonly");
  if (readonly == NULL) 
    fail("could not find readonly");
 
  //locations to match constraints
  location loc6 = malloc(sizeof(struct Location));
  loc6->location_index = 6;

  location loc12 = malloc(sizeof(struct Location));
  loc6->location_index = 12;

  location loc26 = malloc(sizeof(struct Location));
  loc6->location_index = 26;

  //JAVARITESTSUITE.METHODSIG.C <=(6) *C_cons_ret MethodSig.java: 17(313) CLASS
  mkinst_qual(loc6, &open_qgate, METHODSIG_C, cConsRet, p_pos, "test error 1");

  //JAVA.LANG.OBJECT <=(6) Object_cons_ret MethodSig.java: 17(313) CLASS
  mkinst_qual(loc6, &open_qgate, OBJECT, Object_cons_ret, p_pos, "test error 1");

  //JAVARITESTSUITE.METHODSIG <=(12) *MethodSig_cons_ret MethodSig.java: 14(275) CLASS
  mkinst_qual(loc12, &open_qgate, METHODSIG, MethodSig_cons_ret, p_pos, "test error 1");

  //*LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_ret <=(26) LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_invRet@26 MethodSig.java: 29(405) RETURN
  mkinst_qual(loc26, &open_qgate, MethodSig_f0_ret, MethodSig_f0_invRet_26, p_pos, "test error 1");

  //*LjavariTestSuite/MethodSig;.f1()V_this <=(-26) *LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_this MethodSig.java: 29(405) INVOKE
  mkinst_qual(loc26, &open_qgate, MethodSig_f1_this, MethodSig_f0_this, p_neg, "test error 1");

  //LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_invRet@26 <= JAVARITESTSUITE.METHODSIG.C.x<CLASS FIELD> MethodSig.java: 29(410) FIELD
  mkleq_qual(loc, &open_qgate, MethodSig_f0_invRet_26, METHODSIG_C_x, "test error 2");

  //JAVARITESTSUITE.METHODSIG.C.x<CLASS FIELD>_ref <= *mutable MethodSig.java: 29(405) JAVARI
  mkleq_qual(loc, &open_qgate, METHODSIG_C_x_ref, mutable, "test error 2");

  //LjavariTestSuite/MethodSig;.f0()LjavariTestSuite/MethodSig$C;_invRet@26 <= *mutable MethodSig.java: 29(405) JAVARI_MUTABLE
  mkleq_qual(loc, &open_qgate, MethodSig_f0_invRet_26, mutable, "test error 2");


  //lookup readonly
  finish_quals();
  //  fprintf(stderr, "should be zero: %d\n", cs_may_leq_qual(readonly, ret));
  fprintf(stderr, "should be zero: %d\n", may_leq_qual(readonly, MethodSig_f0_ret));
*/

  // pared-down test
  //find qual constants
  qual mutable = find_qual("mutable");
  if (mutable == NULL) 
    fail("could not find mutable");

  qual readonly = find_qual("readonly");
  if (readonly == NULL) 
    fail("could not find readonly");

  qual call = make_qvar("call", loc, TRUE, FALSE);
  set_global_qual (call, TRUE);

  qual ret = make_qvar("ret", loc, TRUE, FALSE);
  qual inv_ret = make_qvar("inv_ret", loc, TRUE, FALSE);

  //call <= ret on (
  //  mkinst_qual(loc, &open_qgate, call, ret, p_neg, "test error 0");  

  //call <= ret
  mkleq_qual(loc, &open_qgate, call, ret, "test error 0");  

  //ret <= inv_ret on )
  mkinst_qual(loc, &open_qgate, ret, inv_ret, p_pos, "test error 1");

  //inv_ret <= mutable
  mkleq_qual(loc, &open_qgate, inv_ret, mutable, "test error 2");
  
  
  //lookup readonly
  finish_quals();
  //  fprintf(stderr, "should be zero: %d\n", cs_may_leq_qual(readonly, ret));
  fprintf(stderr, "from ret: should be zero: %d\n", may_leq_qual(readonly, ret));
  fprintf(stderr, "from call: should be zero: %d\n", may_leq_qual(readonly, call));


}


JNIEXPORT void JNICALL Java_cqual_Native_experiment
(JNIEnv *env, jclass obj) {
  cqual_test();
}
