/* This file is part of cqual.
   Copyright (C) 2003 The Regents of the University of California.

cqual is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

cqual is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with cqual; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */

#include <stdio.h>
#include "cqual.h"
#include "hash.h"
#include "analyze.h"
#include "qerror.h"
#include "qtype.h"
#include "quals.h"
#include "utils.h"
#include "effect.h"

 /* #define DEBUG */
#define STATS

/**************************************************************************
 *                                                                        *
 * Type Declarations                                                      *
 *                                                                        *
 **************************************************************************/

/*
   Constraints
 */

/*  aloc \nin eff  */
typedef struct Nin_constraint
{
  aloc aloc;
  eff_kind k;
  effect eff;
  location loc;

  bool failed;
  bool inference; /* TRUE if this constraint was generated by an
		     inference procedure */
  bool from_forall;
} *nin_constraint;

typedef struct Forall_nin_constraint
{
  eff_kind kind;
  location loc;
  effect eff;

  bool failed;
  bool inference;
} * forall_nin_constraint;


typedef struct Forall_cond_nin_constraint
{
  location loc;
  eff_kind kind1;
  effect eff1;
  eff_kind kind2;
  effect eff2;
  
  bool failed;
  bool inference;
} *forall_cond_nin_constraint;

/*  al1 <=_{i,p} al2 */
typedef struct Instantiation
{
  aloc al1;         
  aloc al2;         
  polarity p;     
  int index;      
} *instantiation;


/*
   Abstract Locations
 */

struct Abstract_loc
{
  enum { aloc_var, aloc_link } kind;
  union
  {
    struct
    {
      const char *name;
      int num_equiv;                 /* Union-find rank */
      effect *effects;               /* Effects constructed from this aloc */
      qtype points_to;               /* The flow-insensitive qtype */
      aloc_set alias_cache;
      bool global:1;                 /* TRUE if this aloc is global */
    } var;
    aloc link;
  } u;
};


/*
  Effects
 */ 
#define MARK_R (1 << (eff_r - 1))
#define MARK_WR (1 << (eff_wr - 1))
#define MARK_ALLOC (1 << (eff_alloc - 1))
#define MARK_RWR (MARK_R | MARK_WR)
#define MARK_ANY (MARK_RWR | MARK_ALLOC)


/* e1 \cap e2 */
typedef struct Inter_info {
  char *name;
  peffect_set reached_1m;         /* reached slot 1 P/N/M */
  peffect_set reached_2m;         /* reached slot 2 P/N/M */
  peffect_set reached_1n;   
  peffect_set reached_2n;   
  peffect_set reached_1p;   
  peffect_set reached_2p;   
  effect slot_1;
  effect slot_2;
  effect e1;
  effect e2;
  effect succ;
} *inter_info;

typedef struct Intersection {
  effect e1;
  effect e2;
  inter_info info;
} *intersection;

struct Effect_edge {
  int index;
  effect bound;
};

typedef enum {pk_match, pk_seed, pk_pos, pk_neg } path_kind;

struct Effect_path {
  path_kind kind;
  effect head;
  effect tail;
  int reached_global;
};

typedef enum { eff_constr, eff_var, eff_union, eff_inter, eff_link } eff_type;

struct Effect
{
  eff_type kind;
 
  union
  {
    struct {
      effect_edge_set m_lbounds;
      effect_edge_set m_ubounds;
      effect_edge_set n_lbounds;
      effect_edge_set n_ubounds;
      effect_edge_set p_lbounds;
      effect_edge_set p_ubounds;

      effect_path_set m_lpath;
      effect_path_set m_upath;
      effect_path_set n_lpath;
      effect_path_set n_upath;
      effect_path_set p_lpath;
      effect_path_set p_upath;

      aloc_int_map reach_cache; 
      bool is_seeded:1;
      bool is_ret:1;
      bool is_param:1;
      bool is_empty:1;
           
      eff_kind kind;
      aloc base;
    } constr;
    struct {
      effect_edge_set m_lbounds;
      effect_edge_set m_ubounds;
      effect_edge_set n_lbounds;
      effect_edge_set n_ubounds;
      effect_edge_set p_lbounds;
      effect_edge_set p_ubounds;

      effect_path_set m_lpath;
      effect_path_set m_upath;
      effect_path_set n_lpath;
      effect_path_set n_upath;
      effect_path_set p_lpath;
      effect_path_set p_upath;
     
      aloc_int_map reach_cache; 
      bool is_seeded:1;
      bool is_ret:1;
      bool is_param:1;
      bool is_empty:1;

      const char *name;
      int num_equiv;
      bool is_inter:1;
      bool reach_failed:1;
    } var;
    inter_info inter_; /* e1 \cap e2, shared */
    struct {
      effect e1;
      effect e2;
    } union_;
    effect link;
  } u;
};


/* Restrict inference
*/

/* the condition k(al) \in e */
typedef struct Rinf_cond {
  bool satisfied;
  aloc al;
  effect e;
  eff_kind k;
} * rinf_cond;

struct Rinf_const {
  location loc;
  struct { /* c1 or c2 or c3 or c4 or c5 => r_al = old_al */
    bool rhs_invoked; 
    rinf_cond c1; /* c1 : r_al \in env */
    rinf_cond c2; /* c2 : r_al \in pt_type */
    rinf_cond c3; /* c3 : r_al \in r_type */
    rinf_cond c4; /* c4 : old_al \in r_body */
    rinf_cond c5; /* c5 : write(top_al) \in r_body */
    aloc r_al;
    aloc old_al;
  } cond1;
  struct { /* c1 => old_al \in body */
    bool rhs_invoked;
    rinf_cond c1; /* c1 : r_al \in r_body */
    aloc old_al;
    effect body;
  } cond2;

};

/* Confine inference 
 */

typedef struct Cinf_const {
  drinfo dri;

  nin_constraint nin1;  /* r-aloc not in tau */
  nin_constraint nin2;  /* r-aloc not in env */
  nin_constraint nin3;  /* old-aloc not in body eff */
  
  forall_nin_constraint forall1; /* no write in confined exp */
  forall_nin_constraint forall2; /* no alloc in confined exp */

  forall_cond_nin_constraint forallcond; /* no write to read aloc */
  int num;
} *cinf_const;

int cinf_cnt = 0;

/**************************************************************************
 *                                                                        *
 * Prototypes                                                             *
 *                                                                        *
 **************************************************************************/

 void extend_effect_paths(effect e1, effect e2);
static unsigned int eff_kind_to_mark(eff_kind k);
static int aloc_in_effect(aloc al, effect goal);
static void unify_constrs(effect *c1, effect *c2);
void internal_mkleq_effect(effect e1, effect e2, polarity p, int index);
static void resolve_instantiations();

#ifdef DEBUG
static int print_effect_path(printf_func pf, effect_path p);
static char *polarity_to_string(polarity p);
static char *eff_kind_to_str(eff_kind k);
#endif

#ifdef STATS
static int rinf_failed_c[5] = {0,0,0,0,0}; 
#endif

/**************************************************************************
 *                                                                        *
 * Global Variables                                                       *
 *                                                                        *
 **************************************************************************/

/* the \emptyset of effects */
effect effect_empty = NULL; 

/* the set of \not\in constraints */
dd_list nin_constraints = NULL; 

/* set of instantiations of alocs   */
dd_list aloc_instantiations = NULL;

/* constraints for restrict inference */
dd_list rinf_constraints = NULL;

/* constraints for confine inference */
dd_list cinf_constraints = NULL;
dd_list forall_constraints = NULL;
dd_list forall_cond_constraints = NULL;


/* permanent region for effects */
region effect_region = NULL;

/* region for resolving instantiations.
   lifetime is from state_init to just before state_reach */
region resolve_region = NULL;

/* a counter for effects */
int next_effect = 0;

/* hash consing for effect reachability */
sig_effect_path_map back_paths;

/* hash consing for alias reachability */
sig_effect_path_map forward_paths;

/* worklist for effect reachability */
dd_list forward_path_worklist = NULL; 

/* worklist for alias reachability */
dd_list back_path_worklist = NULL;

/* state of the module */
static enum { 
  state_raw, state_init, state_reach, state_lin } state = state_raw;

/**************************************************************************
 *                                                                        *
 * Abstract Locations                                                     *
 *                                                                        *
 **************************************************************************/

/* Make a new, unique abstract location */
aloc make_aloc(const char *name, qtype points_to, bool global)
{
  aloc al;

  if (!restrict_qual && !flag_flow_sensitive)
    return NULL;
  
  al = ralloc(effect_region, struct Abstract_loc);
  al->kind = aloc_var;
  al->u.var.name = name;
  al->u.var.num_equiv = 1;
  al->u.var.effects = NULL;
  al->u.var.points_to = points_to;
  al->u.var.global = global;
  al->u.var.alias_cache = empty_aloc_set();
  return al;
}

/* Find equivalence class representative of al */
static aloc ecr_aloc(aloc al)
{
  if (al->kind == aloc_link)
    {
      aloc ecr = al, cur, temp;
      
      /* Find root */
      while (ecr->kind == aloc_link)
	ecr = ecr->u.link;
      
      /* Compress path */
      cur = al;
      while (cur->u.link != ecr)
	{
	  temp = cur->u.link;
	  cur->u.link = ecr;
	  cur = temp;
	}
      return ecr;
    }
  else
    return al;
}

/* Return al's name */
const char *name_aloc(aloc al)
{
  al = ecr_aloc(al);
  assert(al->kind == aloc_var);
  return al->u.var.name;
}

/* Hash function on alocs */
unsigned long hash_aloc(aloc al)
{
  al = ecr_aloc(al);
  return (unsigned long) al;
}

/* Return the qtype al points to */
qtype points_to_aloc(aloc al)
{
  al = ecr_aloc(al);
  assert(al->kind == aloc_var);
  return al->u.var.points_to;
}

bool global_aloc(aloc al)
{
  al = ecr_aloc(al);
  assert(al->kind == aloc_var);
  return al->u.var.global;
}

/* Collect a list of instantiated alocs. Constructing the actual
 instantiation edges is deferred until all the construced effects
 are present.*/
void inst_aloc(location loc, aloc al1, aloc al2, polarity p)
{
  instantiation inst = ralloc(resolve_region, struct Instantiation);
  inst->al1 = al1;
  inst->al2 = al2;
  inst->p = p;
  inst->index = location_index(loc);
  
#ifdef DEBUG
  {
    char *str;
    switch(p)
      {
      case p_pos:
	str = "+";
	break;
      case p_neg:
	str = "-";
	break;
      case p_non:
	str = "T";
	break;
      case p_sub:
	str = "M";
	break;
      default:
	fail("Unexpected polarity");	
      }
    print_aloc(printf, al1);
    printf(" <={%d,%s} ",inst->index,str);
    print_aloc(printf, al2);
    printf("\n");
  }
#endif

  /* add this instantiation edge to the list of instantiations */
  dd_add_last(resolve_region,aloc_instantiations,inst);

  if (state == state_reach)
    resolve_instantiations();
}

/* Add constraint that aloc al1 = al2 */
void unify_aloc(location loc, aloc al1, aloc al2)
{
  if (!restrict_qual && !flag_flow_sensitive)
    return;

#ifdef DEBUG
  print_aloc(printf, al1);
  printf(" %p == ", ecr_aloc(al1));
  print_aloc(printf, al2);
  printf(" %p\n", ecr_aloc(al2));
#endif

  al1 = ecr_aloc(al1);
  al2 = ecr_aloc(al2);

  assert(al1->kind == aloc_var && al2->kind == aloc_var);
  /* How should this go? */
  /*  assert(!al1->u.var.fs_points_to && !al2->u.var.fs_points_to);*/
  if (al1 != al2)
    {
      aloc new_ecr, new_link;

      if (al1->u.var.num_equiv <= al2->u.var.num_equiv)
	{
	  new_ecr = al2;
	  new_link = al1;
	}
      else
	{
	  new_ecr = al1;
	  new_link = al2;
	}

      if (new_ecr->u.var.effects && new_link->u.var.effects)
	unify_constrs(new_ecr->u.var.effects, new_link->u.var.effects);
      else if (new_link->u.var.effects)
	new_ecr->u.var.effects = new_link->u.var.effects;
      
      if (flag_flow_sensitive &&
	  !match_qtype(new_ecr->u.var.points_to, new_link->u.var.points_to,
		       NULL, NULL))
	{
	  report_qerror(loc, sev_err,
			"inconsistent location type for flow-sensitive pass");
	  new_ecr->u.var.points_to = error_qtype;
	}
      
      new_ecr->u.var.global |= new_link->u.var.global;
      new_ecr->u.var.num_equiv += new_link->u.var.num_equiv;
      new_link->kind = aloc_link;      
      new_link->u.link = new_ecr;
   }
}

/* A total ordering on alocs */
int cmp_aloc(aloc left, aloc right)
{
  left = ecr_aloc(left);
  right = ecr_aloc(right);
  return (int) left - (int) right;
}

/* Return TRUE iff al1 and al2 are equal */
bool eq_aloc(aloc al1, aloc al2)
{
  return (ecr_aloc(al1) == ecr_aloc(al2));
}

/* Print the abstract location al */
int print_aloc(printf_func pf, aloc al)
{
  int result;

  al = ecr_aloc(al);
  assert(al->kind == aloc_var);
  result = 0;
  result += pf(al->u.var.name);
  return result;
}

/* Print the abstract location al */
static int print_unique_aloc(printf_func pf, aloc al)
{
  int result;

  al = ecr_aloc(al);
  result = print_aloc(pf, al);
  if (flag_ugly)
    result += pf(" %s", ptr_to_ascii(al));
  return result;
}


/**************************************************************************
 *                                                                        *
 * Effects                                                                *
 *                                                                        *
 **************************************************************************/


void init_effects(void)
{  
  assert(state == state_raw);
  assert(effect_region == NULL);  
  next_effect = 0;

  effect_region = newregion();
  resolve_region = newregion();
  rinf_constraints = dd_new_list(effect_region);
  cinf_constraints = dd_new_list(effect_region);
  forall_constraints = dd_new_list(effect_region);
  forall_cond_constraints = dd_new_list(effect_region);

  nin_constraints = dd_new_list(effect_region);
  aloc_instantiations = dd_new_list(resolve_region);

  forward_paths = make_sig_effect_path_map(effect_region,17);
  back_paths = make_sig_effect_path_map(effect_region,17);

  forward_path_worklist = dd_new_list(effect_region);
  back_path_worklist = dd_new_list(effect_region);

  state = state_init;
}

static effect ecr_effect(effect e)
{
  if (e == NULL || e->kind != eff_link)
    return e;
  else
    {
      effect ecr = e, cur, temp;
        
      /* Find root */
      while (ecr && ecr->kind == eff_link)
	ecr = ecr->u.link;
      
      /* Compress path */
      cur = e;
      while (cur->u.link != ecr)
	{
	  temp = cur->u.link;
	  cur->u.link = ecr;
	  cur = temp;
	}
      return ecr;
    }
}

/* returns aloc from effect */
aloc aloc_effect(effect e)
{
  e = ecr_effect(e);
  assert(e && e->kind == eff_constr);
  return e->u.constr.base;
}

static bool global_effect(effect e)
{
  e = ecr_effect(e);

  if (e->kind == eff_constr)
    return global_aloc(e->u.constr.base);
  else
    return FALSE;
}


static bool seeded_effect(effect e)
{
  e = ecr_effect(e);
  
  if (e->kind == eff_inter)
    return FALSE;

  return e->u.var.is_seeded;
}

static void set_seeded_effect(effect e)
{
  e = ecr_effect(e);

  assert (e->kind != eff_inter);
  
  e->u.var.is_seeded = TRUE;
}

static bool param_effect(effect e)
{
 e = ecr_effect(e);
 
 if (e->kind == eff_inter)
   return FALSE;

 return e->u.var.is_param || global_effect(e);

}

static bool return_effect(effect e)
{
  e = ecr_effect(e);
  
  if (e->kind == eff_inter)
    return FALSE;
  
  return e->u.var.is_ret || global_effect(e);
}

/* mark this aloc as interesting (optimization help for store) */
void mark_aloc_interesting(aloc al)
{
  return;
}

static const char *effect_name(effect e)
{
  const int MAX_SIZE = 512;
  char temp[MAX_SIZE];

  e = ecr_effect(e);
  switch(e->kind)
    {
    case eff_constr:
      switch (e->u.constr.kind)
	{
	case eff_r:
	  snprintf(temp,MAX_SIZE,"read(%s)",
		   ecr_aloc(e->u.constr.base)->u.var.name);
	  break;
	case eff_wr:
	  snprintf(temp,MAX_SIZE,"write(%s)",
		   ecr_aloc(e->u.constr.base)->u.var.name);
	  break;
	case eff_any:
	  snprintf(temp,MAX_SIZE,"any(%s)",
		   ecr_aloc(e->u.constr.base)->u.var.name);
	  break;
	case eff_alloc:
	  snprintf(temp,MAX_SIZE,"alloc(%s)",
		   ecr_aloc(e->u.constr.base)->u.var.name);
	  break;
	case eff_rwr:
	  snprintf(temp,MAX_SIZE,"rwr(%s)",
		   ecr_aloc(e->u.constr.base)->u.var.name);
	  break;
	default:
	  fail("Unexpected constructor kind");
	}
      break;
    case eff_inter:
      return e->u.inter_->name;
      break;
    case eff_var:
      return e->u.var.name;
      break;
    case eff_link:
      return effect_name(ecr_effect(e));
    default:
      break;
    }
  return rstrdup(effect_region,temp);
}


/* The effect k(al) */
effect effect_constr(aloc al, eff_kind k)
{
  int i;
  effect *effects;
  
   /* assert (state == state_init); */
  
  if (!restrict_qual && !flag_flow_sensitive)
    return NULL;

  /* XXX -- hack, eff_any as a union of all effect types */
  if (k == eff_any)
    {
      int i;
      effect result = effect_empty;
      
      for (i = 2; i <= eff_last; i++)
	result = effect_union(result,effect_constr(al,i));

      return result;
    }
  else if (k == eff_rwr)
    {
      return effect_union(effect_constr(al,eff_r),effect_constr(al,eff_wr));
    }

  al = ecr_aloc(al);
  assert(al->kind == aloc_var);
  assert(0 <= k && k <= eff_last);
  
  if (!al->u.var.effects)
    {   
      al->u.var.effects = rarrayalloc(effect_region, eff_last + 1, effect);
      for (i = 0; i <= eff_last; i++)
	al->u.var.effects[i] = NULL;
    }

  effects = al->u.var.effects;
  if (!effects[k]) /* build lists as needed */
    {    
      effect e;
      e = ralloc(effect_region, struct Effect);
      e->kind = eff_constr;
      e->u.constr.base = al;
      e->u.constr.kind = k;
      
      e->u.constr.m_lbounds = empty_effect_edge_set();
      e->u.constr.m_ubounds = empty_effect_edge_set();
      e->u.constr.n_lbounds = empty_effect_edge_set();
      e->u.constr.n_ubounds = empty_effect_edge_set();
      e->u.constr.p_lbounds = empty_effect_edge_set();
      e->u.constr.p_ubounds = empty_effect_edge_set();

      e->u.constr.m_lpath = empty_effect_path_set();
      e->u.constr.m_upath = empty_effect_path_set();
      e->u.constr.n_lpath = empty_effect_path_set();
      e->u.constr.n_upath = empty_effect_path_set();
      e->u.constr.p_lpath = empty_effect_path_set();
      e->u.constr.p_upath = empty_effect_path_set();

      e->u.constr.is_seeded = FALSE;
      e->u.constr.is_ret = FALSE;
      e->u.constr.is_param = FALSE;
      e->u.constr.reach_cache = NULL;
      e->u.constr.is_empty = FALSE;

      effects[k] = e;
    }

  return effects[k];
}

/* The effect {al} */
effect effect_single(aloc al)
{
  return effect_constr(al, eff_any);
}

/* The effect alloc(al), which is different than al itself */
effect effect_alloc(aloc al)
{
  return effect_constr(al, eff_alloc);
}

/* The effect rwr(al) */
effect effect_rwr(aloc al)
{
  return effect_constr(al, eff_rwr);
}

/* The effect r(al) */
effect effect_r(aloc al)
{
  return effect_constr(al, eff_r);
}

/* The effect wr(al) */
effect effect_wr(aloc al)
{
  return effect_constr(al, eff_wr);
}

/* Create a fresh effect variable */
effect effect_var(const char *name)
{
  effect e;

  assert(state == state_init);
  if (!restrict_qual && !flag_flow_sensitive)
    return NULL;

  e = ralloc(effect_region, struct Effect);
  e->kind = eff_var;
  e->u.var.name = name;
  e->u.var.num_equiv = 1;

  e->u.var.m_lbounds = empty_effect_edge_set();
  e->u.var.m_ubounds = empty_effect_edge_set();
  e->u.var.n_lbounds = empty_effect_edge_set();
  e->u.var.n_ubounds = empty_effect_edge_set();
  e->u.var.p_lbounds = empty_effect_edge_set();
  e->u.var.p_ubounds = empty_effect_edge_set();
  
  e->u.var.m_lpath = empty_effect_path_set();
  e->u.var.m_upath = empty_effect_path_set();
  e->u.var.n_lpath = empty_effect_path_set();
  e->u.var.n_upath = empty_effect_path_set();
  e->u.var.p_lpath = empty_effect_path_set();
  e->u.var.p_upath = empty_effect_path_set();
  
  e->u.var.is_seeded = FALSE;
  e->u.var.is_ret = FALSE;
  e->u.var.is_param = FALSE;
  e->u.var.reach_cache = NULL;
  e->u.var.is_inter = FALSE;
  e->u.var.reach_failed = FALSE;
  e->u.var.is_empty = TRUE;

  return e;
}

/* Create a fresh effect variable with a unique name */
effect effect_fresh(void)
{
  const char *name;

  if (!restrict_qual && !flag_flow_sensitive)
    return NULL;

  name = rstrcat(effect_region, "ef",
		 inttostr(effect_region, next_effect++));
  return effect_var(name);
}

/* The effect e1 \cup e2 */
effect effect_union(effect e1, effect e2)
{
  if (!restrict_qual && !flag_flow_sensitive)
    return NULL;

  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);

  if (e1 == e2)
    return e1;

  if (e1 == NULL)
    return e2;
  else if (e2 == NULL)
    return e1;

  else
    { 
      effect e;

      e = ralloc(effect_region, struct Effect);
      e->kind = eff_union;
      e->u.union_.e1 = e1;
      e->u.union_.e2 = e2;

      return e;
    }
}

/* The effect e1 \cap e2.  e2 cannot be the empty effect. */
effect effect_inter(effect e1, effect e2)
{

  if (!restrict_qual && !flag_flow_sensitive)
    return NULL;
  
  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);

  assert(e2 != NULL);
  
  if (e1 == NULL)
    return effect_empty;
  
  else if (eq_effect(e1,e2))
    return e1;

  
  else
    {
      effect e, inter_1, inter_2;
      inter_info info;
      
      assert(e2 != NULL);
  
      e = effect_fresh();
      e->u.var.is_inter = TRUE;

#ifdef DEBUG
      printf("[");
      print_effect(printf,e1);
      printf(" & ");
      print_effect(printf,e2);
      printf("] <= ");
      print_effect(printf,e);
      printf("\n");
#endif
      inter_1 = ralloc(effect_region, struct Effect);
      inter_2 = ralloc(effect_region, struct Effect);
      info = ralloc(effect_region, struct Inter_info);
      
      inter_1->kind = eff_inter;
      inter_2->kind = eff_inter;
      inter_1->u.inter_ = info;
      inter_2->u.inter_ = info;
           
      info->name = rstrcat(effect_region,"inter_",
			   inttostr(effect_region, next_effect++));
      info->slot_1 = inter_1;
      info->slot_2 = inter_2;
      info->e1 = e1;
      info->e2 = e2;
      info->succ = e;
      
      info->reached_1m =
	empty_peffect_set();

      info->reached_2m =
	empty_peffect_set();

      info->reached_1n =
	empty_peffect_set();

      info->reached_2n =
	empty_peffect_set();

      info->reached_1p =
	empty_peffect_set();

      info->reached_2p =
	empty_peffect_set();

      internal_mkleq_effect(e1,inter_1,p_sub,0);
      internal_mkleq_effect(e2,inter_2,p_sub,0);

      return e;
    }

}

/* Mark this effect as interesting (optimization for caching
   reachability queries) */
void mark_effect_interesting(effect e)
{
  return;
}
/* Return FALSE if e1 may not be equal to e2.  Does not generate a
   constraint. */
bool eq_effect(effect e1, effect e2)
{
  return (ecr_effect(e1) == ecr_effect(e2));
}

/* A total ordering on effects */
int cmp_effect(effect left, effect right)
{
  left = ecr_effect(left);
  right = ecr_effect(right);
  return (int) left - (int) right;
}

/* A total ordering on effect paths */
int cmp_effect_path(effect_path left, effect_path right)
{
  if (left->kind != right->kind)
    return (left->kind - right->kind);
  else if (! eq_effect(left->head,right->head))
    return cmp_effect(left->head,right->head);
  else if (! eq_effect(left->tail,right->tail))
    return cmp_effect(left->tail,right->tail);
  
  return (int)left->reached_global - (int)right->reached_global;
}

/* A total ordering on effect edges */
int cmp_effect_edge(effect_edge left, effect_edge right)
{
  if (! eq_effect(left->bound,right->bound))
    return cmp_effect(left->bound,right->bound);
  
  return (int)left->index - (int)right->index;
}

void internal_mkleq_effect(effect e1, effect e2, polarity p, int index)
{
  effect_edge ub,lb;
  
  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);

  if (e1 == NULL)
    return;

   /* assert(e2->kind == eff_var); */

  assert ( (e1->kind != eff_inter) );
  assert ( (e2->kind != eff_union) );

  if (e1->kind == eff_union)
    {
      internal_mkleq_effect(e1->u.union_.e1,e2,p,index);
      internal_mkleq_effect(e1->u.union_.e2,e2,p,index);
    }

  else if (e2->kind == eff_inter)
    {
      assert(state == state_init);
      assert(e1->kind == eff_constr || e1->kind == eff_var);

      ub = ralloc(effect_region, struct Effect_edge);
      ub->index = index;
      ub->bound = e2;

      switch(p)
	{
	case p_pos:
	  e1->u.constr.is_ret = TRUE;
	  effect_edge_set_insert(effect_region,&e1->u.constr.p_ubounds,ub);
	  break;
	case p_neg:
	  e1->u.constr.is_param = TRUE;
	  effect_edge_set_insert(effect_region,&e1->u.constr.n_ubounds,ub);
	  break;
	case p_sub:
	  effect_edge_set_insert(effect_region,&e1->u.constr.m_ubounds,ub);
	  break;
	case p_non: /* XXX - fix! */
	  /* HACK -- one level of polymorphism only */
	  effect_edge_set_insert(effect_region,&e1->u.constr.m_ubounds,ub);
	   /* effect_edge_set_insert(effect_region,&e1->u.constr.m_ubounds,ub); */
	  break;
	default : 
	  fail("Unexpected polarity %x\n",p);
	}

    }
  else
    {
      
      assert(e1->kind == eff_constr || e1->kind == eff_var);
      assert(e2->kind == eff_constr || e2->kind == eff_var);


      if (e2->kind == eff_var)
	{
	  assert(e2->u.var.is_inter == FALSE);
	}

      ub = ralloc(effect_region, struct Effect_edge);
      lb = ralloc(effect_region, struct Effect_edge);
      
      ub->index = index;
      ub->bound = e2;
      lb->index = index;
      lb->bound = e1;
      
      switch(p)
	{
	case p_pos:
	  e1->u.constr.is_ret = TRUE;
	  effect_edge_set_insert(effect_region,&e1->u.constr.p_ubounds,ub);
	  effect_edge_set_insert(effect_region,&e2->u.constr.p_lbounds,lb);
	  break;
	case p_neg:
	  e1->u.constr.is_param = TRUE;
	  effect_edge_set_insert(effect_region,&e1->u.constr.n_ubounds,ub);
	  effect_edge_set_insert(effect_region,&e2->u.constr.n_lbounds,lb);
	  break;
	case p_sub:
	  effect_edge_set_insert(effect_region,&e1->u.constr.m_ubounds,ub);
	  effect_edge_set_insert(effect_region,&e2->u.constr.m_lbounds,lb);
	  break;
	case p_non: /* XXX - fix! */
	  /* HACK -- one level of polymorphism only */
	  effect_edge_set_insert(effect_region,&e1->u.constr.m_ubounds,ub);
	  effect_edge_set_insert(effect_region,&e2->u.constr.m_lbounds,lb);
	  effect_edge_set_insert(effect_region,&e1->u.constr.m_lbounds,lb);
	  effect_edge_set_insert(effect_region,&e2->u.constr.m_ubounds,ub);
	  break;
	default : 
	  fail("Unexpected polarity %x\n",p);
	}

      if (state == state_reach)
	{
	  assert (p == p_sub);
	  extend_effect_paths(e1,e2);
	}
      
    }
}



/* Add constraint that effect set e1 is a subset of effect set e2. */
void mkleq_effect(effect e1, effect e2)
{
#ifdef DEBUG
  print_effect(printf, e1);
  printf(" <= ");
  print_effect(printf, e2);
  printf("\n");
#endif
  internal_mkleq_effect(e1,e2,p_sub,0);
}

/* Add instantiation constraint e1 \preceq_{p,i} e2 */
void mkinst_effect(location l,effect e1, effect e2, polarity p)
{
#ifdef DEBUG
  print_effect(printf, e1);
  printf(" <={%s,%d} ",polarity_to_string(p),location_index(l));
  print_effect(printf, e2);
  printf("\n");
#endif

  internal_mkleq_effect(e1,e2,p,location_index(l));
}

/* Add constraint that effect set e1 is equal to the effect set e2. */
void mkeq_effect(effect e1, effect e2) 
{
#ifdef DEBUG
  print_effect(printf, e1);
  printf(" = ");
  print_effect(printf, e2);
  printf("\n");
#endif

  internal_mkleq_effect(e1,e2,p_sub,0);
  internal_mkleq_effect(e2,e1,p_sub,0);
}

/* Add constraint that effect sets e1 and e2 are equal and
   indistinguishable.  e1 and e2 must be effect variables. */
void unify_effect(effect e1, effect e2)
{

  effect new_ecr, new_link;

  if (!restrict_qual && !flag_flow_sensitive)
    return;

#ifdef DEBUG
  print_effect(printf, e1);
  printf(" == ");
  print_effect(printf, e2);
  printf("\n");
#endif


  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);

  if (eq_effect(e1,e2))
    return;

  assert(e1->kind == eff_var && e2->kind == eff_var);
  assert(state == state_init);
  assert( (!e1->u.var.is_inter) && (! e2->u.var.is_inter));

  if (e1->u.var.num_equiv <= e2->u.var.num_equiv)
    {
      new_ecr = e2;
      new_link = e1;
    }
  else
    {
      new_ecr = e1;
      new_link = e2;
    }

  new_ecr->u.var.num_equiv += new_link->u.var.num_equiv;

  new_ecr->u.var.m_lbounds = effect_edge_set_union(new_ecr->u.var.m_lbounds,
					    new_link->u.var.m_lbounds);
  new_ecr->u.var.m_ubounds = effect_edge_set_union(new_ecr->u.var.m_ubounds,
					    new_link->u.var.m_ubounds);

  new_ecr->u.var.n_lbounds = effect_edge_set_union(new_ecr->u.var.n_lbounds,
					    new_link->u.var.n_lbounds);
  new_ecr->u.var.n_ubounds = effect_edge_set_union(new_ecr->u.var.n_ubounds,
					    new_link->u.var.n_ubounds);

  new_ecr->u.var.p_lbounds = effect_edge_set_union(new_ecr->u.var.p_lbounds,
					    new_link->u.var.p_lbounds);
  new_ecr->u.var.p_ubounds = effect_edge_set_union(new_ecr->u.var.p_ubounds,
					    new_link->u.var.p_ubounds);
  
 
  assert(effect_path_set_empty(new_ecr->u.var.m_lpath));
  assert(effect_path_set_empty(new_ecr->u.var.m_upath));
  assert(effect_path_set_empty(new_ecr->u.var.p_lpath));
  assert(effect_path_set_empty(new_ecr->u.var.p_upath));
  assert(effect_path_set_empty(new_ecr->u.var.n_lpath));
  assert(effect_path_set_empty(new_ecr->u.var.n_upath));
  assert(new_ecr->u.var.is_seeded == FALSE);

  new_ecr->u.var.is_ret |= new_link->u.var.is_ret;
  new_ecr->u.var.is_param |= new_link->u.var.is_param;
  new_ecr->u.var.is_empty |= new_link->u.var.is_empty;
  new_ecr->u.var.reach_failed |= new_link->u.var.reach_failed;
  

  new_link->kind = eff_link;
  new_link->u.link = new_ecr;
}


/* Unify the constructed effects in c1 and c2 */
static void unify_constrs(effect *c1, effect *c2)
{
  int i;
  
  for (i = 0; i <= eff_last; i++)
    {
      effect e1, e2;

      e1 = ecr_effect(c1[i]);
      e2 = ecr_effect(c2[i]);
      if (!e1)
	c1[i] = e2;
      else if (!e2)
	c2[i] = e1;
      else
	{
	  assert(e1->kind == eff_constr && e2->kind == eff_constr &&
		 e1->u.constr.kind == e2->u.constr.kind);
	  if (e1 != e2)
	    {
	      e1->u.var.m_lbounds = effect_edge_set_union(e1->u.var.m_lbounds,
							  e2->u.var.m_lbounds);
	      e1->u.var.m_ubounds = effect_edge_set_union(e1->u.var.m_ubounds,
							  e2->u.var.m_ubounds);
	      
	      e1->u.var.n_lbounds = effect_edge_set_union(e1->u.var.n_lbounds,
							  e2->u.var.n_lbounds);
	      e1->u.var.n_ubounds = effect_edge_set_union(e1->u.var.n_ubounds,
							  e2->u.var.n_ubounds);
	      
	      e1->u.var.p_lbounds = effect_edge_set_union(e1->u.var.p_lbounds,
							  e2->u.var.p_lbounds);
	      e1->u.var.p_ubounds = effect_edge_set_union(e1->u.var.p_ubounds,
							  e2->u.var.p_ubounds);
	      
	      /* these assertions will fail if we unify alocs during
		 reachability (as for restrict inference) */
	      assert(effect_path_set_empty(e1->u.var.m_lpath));
	      assert(effect_path_set_empty(e1->u.var.m_upath));
	      assert(effect_path_set_empty(e1->u.var.p_lpath));
	      assert(effect_path_set_empty(e1->u.var.p_upath));
	      assert(effect_path_set_empty(e1->u.var.n_lpath));
	      assert(effect_path_set_empty(e1->u.var.n_upath));
	      assert(e1->u.var.is_seeded == FALSE);
	      
	      e1->u.var.is_ret |= e2->u.var.is_ret;
	      e1->u.var.is_param |= e2->u.var.is_param;
	      
	      e2->kind = eff_link;
	      e2->u.link = e1;
	    }
	}
    }
}

nin_constraint internal_mknin_aloc_effect(location loc, aloc al, eff_kind kind, effect e)
{
 nin_constraint c;
  
  e = ecr_effect(e);
  
  if (e == NULL)
    return NULL;
  else if (e->kind != eff_var)
    {
      effect temp;
      
      temp = effect_fresh();
      mkleq_effect(e,temp);
      e = temp;
    }

  c = ralloc(effect_region, struct Nin_constraint);
  c->aloc = al;
  c->eff = e;
  c->loc = loc;
  c->k = kind;
  
  c->failed = FALSE;
  c->inference = FALSE;
  c->from_forall = FALSE;

   /* constraint_list_cons(c,constraints);   */
  dd_add_last(effect_region,nin_constraints,c);
  return c;
}

/* Remember the constraint al !<= e2.  The constraint will be checked
   when check_nin is called. */
void mknin_aloc_effect(location loc, aloc al, eff_kind kind, effect e)
{
  internal_mknin_aloc_effect(loc,al,kind,e);
}

/* Remember the constraint e1 !<= e2.  The constraint will be checked
  when check_nin is called.  e1 must not contain any effect variables.  */
void mknin_effect(location loc, effect e1, effect e2)
{
  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);
  
  if (e1 == NULL || e2 == NULL)     /* Anything is <= !0 and 0 <= anything */
    return;

  if (e1->kind == eff_constr)
    mknin_aloc_effect(loc, e1->u.constr.base,e1->u.constr.kind, e2);
  else if (e1->kind == eff_union)
    {
      mknin_effect(loc, e1->u.union_.e1, e2);
      mknin_effect(loc, e1->u.union_.e2, e2);
    }
  else
    fail("Unexpected effect kind %x\n", e1->kind);
}

static effect normalize_effect(effect e)
{
  if (e == NULL)  /* nothing is in 0 */
    return NULL;
  else if (e->kind != eff_var)
    {
      effect temp;
      temp = effect_fresh();
      mkleq_effect(e, temp);
      e = temp;
    }
  return e;
}

/* Make the lhs of a conditional constraint */
static rinf_cond mkrinf_cond(aloc al, effect e, eff_kind k)
{
  rinf_cond new_cond;
  
  new_cond = ralloc(effect_region, struct Rinf_cond);
  new_cond->satisfied = 0;
  new_cond->al = al;
  new_cond->e = e;
  new_cond->k = k;
  
  return new_cond;
}

rinf_const mkrinf_constraint(location loc,
		       aloc r_al, aloc old_al,
		       aloc top_al,
		       effect r_type, effect pt_type,
		       effect r_body, effect env)
{
  rinf_const rc;
  
  r_type = normalize_effect(r_type);
  pt_type = normalize_effect(pt_type);
  r_body = normalize_effect(r_body);
  env = normalize_effect(env);

  rc = ralloc(effect_region, struct Rinf_const);

#ifdef DEBUG
  printf("-------------------------------------\n");
  printf("Mk_rinf_const:\n");
  printf("Restricted location: ");
  print_aloc(printf, r_al);
  printf("\n");
  printf("Old location: ");
  print_aloc(printf, old_al);
  printf("\n");
  printf("Topmost location: ");
  print_aloc(printf, top_al);
  printf("\n");
  printf("Return type effect: ");
  print_effect(printf, r_type);
  printf("\n");
  printf("Points-to type effect: ");
  print_effect(printf, pt_type);
  printf("\n");
  printf("Restricted body effect: ");
  print_effect(printf, r_body);
  printf("\n");

  printf("Environment: ");
  print_effect(printf, env);
  printf("\n");
#endif


  rc->loc = loc;
  rc->cond1.c1 = 
    mkrinf_cond(r_al, env,eff_any);
  rc->cond1.c2 = 
    mkrinf_cond(r_al, pt_type,eff_any);
  rc->cond1.c3 = 
    mkrinf_cond(r_al, r_type,eff_any);
  rc->cond1.c4 = 
    mkrinf_cond(old_al, r_body,eff_any);
  rc->cond1.c5 = 
    mkrinf_cond(top_al, r_body,eff_wr);
  rc->cond1.r_al = r_al;
  rc->cond1.old_al = old_al;

  rc->cond2.c1 =
    mkrinf_cond(r_al, r_body,eff_any);
  rc->cond2.old_al = old_al;

  rc->cond1.rhs_invoked = 0;
  rc->cond2.rhs_invoked = 0;

   /* DEBUG */
  dd_add_last(effect_region, rinf_constraints, rc);

  return rc;
}

void set_bodyeffect(rinf_const rc, effect body)
{
  body = normalize_effect(body);
  rc->cond2.body = body;

#ifdef DEBUG
  printf("Body effect: ");
  print_effect(printf, body);
  printf("\n");
  printf("-------------------------------------\n");
#endif
}


bool rinf_satisfied(rinf_const c)
{

  if (!c->cond1.rhs_invoked)
    return TRUE;
  else
    {
#ifdef STATS
      if (c->cond1.c1->satisfied)
	  rinf_failed_c[0]++;
      if (c->cond1.c2->satisfied)
	  rinf_failed_c[1]++;
      if (c->cond1.c3->satisfied)
	  rinf_failed_c[2]++;
      if (c->cond1.c4->satisfied)
	  rinf_failed_c[3]++;
      if (c->cond1.c5->satisfied)
	  rinf_failed_c[4]++;
#endif      
      return FALSE;
    }
}

void check_rinf_constraints(void)
{
  int iter = 0;
  bool some_rhs_invoked = FALSE;
  dd_list_pos cur;
  region scratch;
  
  assert(state == state_reach);

   
#ifdef DEBUG
  printf("%s","Computing constraints...\n");
#endif  

  do
    {
      dd_list conditions;
          
      scratch = newregion();
      conditions = dd_new_list(scratch);


      printf("%s : %d\n","Iteration", ++iter);

  
      /* Build up a worklist of conditionals to check */
      dd_scan(cur,rinf_constraints)
      { 
	rinf_const c = DD_GET(rinf_const,cur);
	
	if (!c->cond1.rhs_invoked)
	  {
	    dd_add_last(scratch,conditions,c->cond1.c1);
	    dd_add_last(scratch,conditions,c->cond1.c2);
	    dd_add_last(scratch,conditions,c->cond1.c3);
	    dd_add_last(scratch,conditions,c->cond1.c4);
	    dd_add_last(scratch,conditions,c->cond1.c5);
	  }
	
	if (!c->cond2.rhs_invoked)
	  dd_add_last(scratch,conditions,c->cond2.c1);
      }
      
      /* Compute reachability for each conditional */
      dd_scan(cur,conditions)
	{
	  rinf_cond cond;
	  effect e;
	  aloc al;
	  
	  cond = DD_GET(rinf_cond,cur);
	  e = ecr_effect(cond->e);
	  al = ecr_aloc(cond->al);

#ifdef DEBUG
	  print_aloc(printf,al);
	  printf(" -?-> ");
	  print_effect(printf,e);
	  printf("\n");
#endif	  

	  if (e)
	    cond->satisfied = 
	      aloc_in_effect(al,e) & eff_kind_to_mark(cond->k);

#ifdef DEBUG
	  print_aloc(printf,al);
	  printf(" -?-> ");
	  print_effect(printf,e);
	  if (cond->satisfied)
	    printf(" (SATISFIED)");
	  printf("\n");
#endif
	}

      /* Fire off rhs as necessary */
      some_rhs_invoked = FALSE;
      dd_scan(cur, rinf_constraints)
	{
	  rinf_const c = DD_GET(rinf_const, cur);
	  
	  if (!c->cond1.rhs_invoked &&
	      (c->cond1.c1->satisfied || c->cond1.c2->satisfied ||
	       c->cond1.c3->satisfied || c->cond1.c4->satisfied ||
	       c->cond1.c5->satisfied)
	      )
	    {
	      inst_aloc(c->loc,c->cond1.r_al,c->cond1.old_al,p_sub);
	      inst_aloc(c->loc,c->cond1.old_al,c->cond1.r_al,p_sub);
	      c->cond1.rhs_invoked = some_rhs_invoked = TRUE;
	    }

	  if (!c->cond2.rhs_invoked && c->cond2.c1->satisfied)
	    {
	      effect eff = effect_single(c->cond2.old_al);

#ifdef DEBUG
	  print_aloc(printf,c->cond2.c1->al);
	  printf(" -?-> ");
	  print_effect(printf,c->cond2.c1->e);
	  if (c->cond2.c1->satisfied)
	    printf(" (FIRING RHS)");
	  printf("\n");
#endif

	      mkleq_effect(eff,c->cond2.body);
	      c->cond2.rhs_invoked = some_rhs_invoked = TRUE;
	    }
	}

      if (some_rhs_invoked)
	{
	   /*   invalidate_aloc_reach_caches(); */
	}
      deleteregion(scratch);
  }
  while(some_rhs_invoked);
  
  /* Mark inferred restricts */
  {
    int inferred, total;
    
    inferred = total = 0;
    
    dd_scan(cur, rinf_constraints)
      {
	rinf_const c = DD_GET(rinf_const, cur);

	if (rinf_satisfied(c))
	  {
	    inferred++;
	  }
	
	total++;
      }
    
    printf("Restrict Inferred: %d/%d\n", inferred, total);
#ifdef STATS
    printf("Failed Inference Conditions: \n");
    printf("Escaped (environment) : %d\n",rinf_failed_c[0]);
    printf("Escaped (points-to type) : %d\n",rinf_failed_c[1]);
    printf("Escaped (return type) : %d\n",rinf_failed_c[2]);
    printf("Old location accessed : %d\n",rinf_failed_c[3]);
    printf("Pointer written : %d\n",rinf_failed_c[4]);
#endif
  }

}

static forall_nin_constraint internal_forall_aloc_mknin_effect(location loc, eff_kind kind, effect e)
{
  forall_nin_constraint c;
  
  e = ecr_effect(e);
  
  if (e == NULL)
    return NULL;
  else if (e->kind != eff_var)
    {
      effect temp;

      temp = effect_fresh();
      mkleq_effect(e, temp);
      e = temp;
    }
  
  mark_effect_interesting(e);
  
  c = ralloc(effect_region, struct Forall_nin_constraint);
  c->kind = kind;
  c->loc = loc;
  c->eff = e;
  c->failed = FALSE;
  c->inference = FALSE;
  dd_add_last(effect_region, forall_constraints, c);
  
  return c;
}

void forall_aloc_mknin_effect(location loc, eff_kind k, effect e)
{
  internal_forall_aloc_mknin_effect(loc,k,e);
  return;
}

static forall_cond_nin_constraint internal_forall_aloc_in_effect_mknin_effect(location loc, eff_kind kind1, effect e1, eff_kind kind2, effect e2)
{
  forall_cond_nin_constraint c;

  e1 = ecr_effect(e1);
  
  if (e1 == NULL)
    return NULL;
  else if (e1->kind != eff_var)
    {
      effect temp;

      temp = effect_fresh();
      mkleq_effect(e1, temp);
      e1 = temp;
    }
  
  mark_effect_interesting(e1);

  if (e2 == NULL)
    return NULL;
  else if (e2->kind != eff_var)
    {
      effect temp;

      temp = effect_fresh();
      mkleq_effect(e2, temp);
      e2 = temp;
    }
  
  mark_effect_interesting(e2);

  c = ralloc(effect_region, struct Forall_cond_nin_constraint);
  c->kind1 = kind1;
  c->kind2 = kind2;
  c->loc = loc;
  c->eff1 = e1;
  c->eff2 = e2;
  c->failed = FALSE;
  c->inference = FALSE;
  dd_add_last(effect_region, forall_cond_constraints, c);

  return c;
}

void forall_aloc_in_effect_mknin_effect(location loc,
					eff_kind kind1, effect e1,
					eff_kind kind2, effect e2)
{
  internal_forall_aloc_in_effect_mknin_effect(loc,kind1,e1,kind2,e2);
}

/* XXX -- take out of interface, too */
bool check_nin_aloc_effect(aloc al, effect e)
{
  assert(0);
  return FALSE;
}

void mk_confine_inf_const(drinfo dri)
{
  cinf_const cc = ralloc(effect_region, struct Cinf_const);

  cinf_cnt++;
  cc->num = cinf_cnt;

  assert(dri);
  cc->dri = dri;
  cc->nin1 = 
    internal_mknin_aloc_effect(dri->location, 
			       aloc_qtype(dri->rqtype),
			       eff_rwr, 
			       alocs_qtype(points_to_qtype(dri->qtype)));
  if (cc->nin1)
    cc->nin1->inference = TRUE;
  cc->nin2 =
    internal_mknin_aloc_effect(dri->location,
			       aloc_qtype(dri->rqtype),
			       eff_any,
			       dri->env);
  if (cc->nin2)
    cc->nin2->inference = TRUE;
  cc->nin3 =
    internal_mknin_aloc_effect(dri->location,
			       aloc_qtype(dri->qtype),
			       eff_rwr,
			       dri->body_eff);

  if (cc->nin3)
    cc->nin3->inference = TRUE;
  cc->forall1 =
    internal_forall_aloc_mknin_effect(dri->location,
				      eff_wr,
				      dri->effect);
  if (cc->forall1)
      cc->forall1->inference = TRUE;
  cc->forall2 =
    internal_forall_aloc_mknin_effect(dri->location,
				      eff_alloc,
				      dri->effect);
  if (cc->forall2)
    cc->forall2->inference = TRUE;
  cc->forallcond = 
    internal_forall_aloc_in_effect_mknin_effect(dri->location,
						eff_r,
						dri->effect,
						eff_wr,
						dri->body_eff);
  if (cc->forallcond)
    cc->forallcond->inference = TRUE;
  dd_add_last(effect_region, cinf_constraints, cc);
}

void check_effect_constraints(void)
{
  dd_list_pos cur;
  bool some_rhs_invoked;
  int failed_count = 0;

  assert(state == state_init);

  resolve_instantiations();
  
  state = state_reach;


  do {
    check_nin();
    some_rhs_invoked = FALSE;

    dd_scan(cur,cinf_constraints)
     {
	cinf_const cc = DD_GET(cinf_const, cur);
	if (!cc->dri->failed &&
	    ((cc->nin1 && cc->nin1->failed) || 
	     (cc->nin2 && cc->nin2->failed) || 
	     (cc->nin3 && cc->nin3->failed) ||
	     (cc->forall1 && cc->forall1->failed) || 
	     (cc->forall2 && cc->forall2->failed) ||
	     (cc->forallcond && cc->forallcond->failed)))
	  {
	    assert(cc->dri->inference);
	    cc->dri->failed = TRUE;
	    inst_aloc(cc->dri->location, 
		       aloc_qtype(cc->dri->rqtype), 
		       aloc_qtype(cc->dri->qtype),p_sub);
	    inst_aloc(cc->dri->location, 
		      aloc_qtype(cc->dri->qtype),
		      aloc_qtype(cc->dri->rqtype),
		      p_sub);
	    some_rhs_invoked = TRUE;
	    failed_count++;
	    /* printf("Failed #%d\n", cc->num); */
	  }
     }
  } while (some_rhs_invoked);
  
}

/**************************************************************************
 *                                                                        *
 * Reachability                                                           *
 *                                                                        *
 **************************************************************************/

static void resolve_instantiations()
{
  bool new_constr_made = TRUE;

  dd_list_pos cur;
  instantiation inst;
  int i;

do
  {
    new_constr_made = FALSE;
    
    dd_scan(cur,aloc_instantiations)
      {
	aloc al1,al2;
	
	inst = DD_GET(instantiation,cur);
	
	al1 = ecr_aloc(inst->al1);
	al2 = ecr_aloc(inst->al2);
	
	if (!al1->u.var.effects && !al2->u.var.effects)
	  continue;
	
	if (!al1->u.var.effects)
	  {
	    al1->u.var.effects = 
	      rarrayalloc(effect_region, eff_last + 1, effect);
	    for (i = 0; i <= eff_last; i++)
	      al1->u.var.effects[i] = NULL;
	  }
	
	else if (!al2->u.var.effects)
	  {
	    al2->u.var.effects = 
	      rarrayalloc(effect_region, eff_last + 1, effect);
	    for (i = 0; i <= eff_last; i++)
	      al2->u.var.effects[i] = NULL;
	  }
	
	for (i = 0; i <= eff_last;i++)
	  {
	    if (al1->u.var.effects[i] && !al2->u.var.effects[i])
	      {
		al2->u.var.effects[i] = effect_constr(al2,i);
		new_constr_made = TRUE;
	      }
	    else if ( !al1->u.var.effects[i] && al2->u.var.effects[i])
	      {
		al1->u.var.effects[i] = effect_constr(al1,i);
		new_constr_made = TRUE;
	      }
	    if (al1->u.var.effects[i] && al2->u.var.effects[i])
	      internal_mkleq_effect(al1->u.var.effects[i],
				    al2->u.var.effects[i],inst->p,
				    inst->index);
	  }
      }
  }
 while(new_constr_made); 
 deleteregion(resolve_region);
 resolve_region = newregion();
 aloc_instantiations = dd_new_list(resolve_region);
}


static unsigned int eff_kind_to_mark(eff_kind k)
{
  unsigned int mask;
  switch (k)
    {
    case eff_any:
      mask = MARK_ANY;
      break;
    case eff_rwr:
      mask = MARK_RWR;
      break;
    case eff_r:
      mask = MARK_R;
      break;
    case eff_wr:
      mask = MARK_WR;
      break;
    case eff_alloc:
      mask = MARK_ALLOC;
      break;
    default:
      fail("Unexpected constraint kind %x\n", k);
    }
  return mask;
}

/*
static char *eff_type_to_str(eff_type t)
{     
  switch (t)
    {
    case eff_constr:
      return "constr";
      break;
    case eff_union:
      return "union";
      break;
    case eff_inter:
      return "inter";
      break;
    case eff_var:
      return "var";
      break;
    case eff_link:
      return "link";
      break;
    default:
      fail("Unexpected effect type %x\n", t);
    }
}
*/

unsigned long path_sig_hash(int *sig)
{
  int i;
  unsigned long h = 0;

  for (i = 0; i < 4; i++)
    {
      h = 33*h + 720 + sig[i];
    }
  return h;
}

bool eq_path_sig(int *p1, int *p2)
{
  int i;
  for (i = 0; i < 4; i++)
    {
      if (p1[i] != p2[i])
	return FALSE;
    }
  return TRUE;
}


/* create a new path for either forward or backward reachability, and
   add it to the appropriate worklist. return FALSE if reachability
   failed (e.g., backwards reaches an intersection) */
static bool internal_mkeffect_path(path_kind k, effect head, 
				   effect tail,bool reached_global,
				   bool back_reachability)
{
  effect_path result;
  dd_list worklist;

  sig_effect_path_map hash;

  region scratch_region = newregion();
  int *sig = rarrayalloc(scratch_region,4,int);
  sig[0] = (int)k;
  sig[1] = (int)ecr_effect(head);
  sig[2] = (int)ecr_effect(tail);
  sig[3] = (int)reached_global;

  head = ecr_effect(head);
  tail = ecr_effect(tail);

  assert(head->kind != eff_union && tail->kind != eff_union);

  /* Select the correct worklist and hash for hash consing */
  if (back_reachability)
    {
      hash = back_paths;
      worklist = back_path_worklist;

      /* Backwards reachability fails through intersections */
      if ( (head->kind == eff_var && head->u.var.is_inter) )
	{
	  head->u.var.reach_failed = TRUE;
	  
	  if (tail->kind == eff_var)
	    tail->u.var.reach_failed = TRUE;

	  return FALSE;
	}
      if ( (tail->kind == eff_var && tail->u.var.is_inter) )
	{
	  tail->u.var.reach_failed = TRUE;
	  return FALSE;
	}
    }
  else
    {
      hash = forward_paths;
      worklist = forward_path_worklist;
    }

  /* Cache any aliases if both effects are constructed effects */
  if (head->kind == eff_constr && tail->kind == eff_constr)
    {
      aloc al,alias;
      al = ecr_aloc(tail->u.constr.base);
      alias = ecr_aloc(head->u.constr.base);
      aloc_set_insert(effect_region,&al->u.var.alias_cache,alias);
    }
  
  /* Cache the fact that an aloc reached an effect */
  if (head->kind == eff_constr && tail->kind != eff_inter)
    {
      unsigned int mark;
      if (!tail->u.var.reach_cache)
	tail->u.var.reach_cache = make_aloc_int_map(effect_region,17);

      aloc_int_map_lookup(tail->u.var.reach_cache,
			  head->u.constr.base,&mark);
      
      mark |= eff_kind_to_mark(head->u.constr.kind);
      
      aloc_int_map_insert(tail->u.var.reach_cache,head->u.constr.base,
			  mark);

      tail->u.var.is_empty = FALSE;
      
#ifdef DEBUG
      printf("An aloc reached an effect : ");
      print_aloc(printf,head->u.constr.base);
      printf(" into ");
      print_effect(printf,tail);
      printf("\n");
#endif
    }
  
  /* Hash consing. If a path with the same signature is not present, add it
     to the worklist and the hash table
   */
  if (! sig_effect_path_map_lookup(hash,sig,&result))
    {
      int *psig;
      result = ralloc(effect_region,struct Effect_path);
      result->head = head;
      result->tail = tail;
      result->kind = k;
      result->reached_global = reached_global;
      
      psig = rarrayalloc(effect_region,4,int);
      rarraycopy(psig,sig,4, int);


#ifdef DEBUG
      printf("Discovered path : ");
      print_effect_path(printf,result);
      printf("\n");
#endif

      assert(head->kind != eff_inter);

      sig_effect_path_map_insert(hash,psig,result);
      dd_add_last(effect_region,worklist,result);
      
      switch(k)
	{
	case pk_pos:
	  effect_path_set_insert(effect_region,&head->u.var.p_upath,result);
	  if (tail->kind != eff_inter)
	    effect_path_set_insert(effect_region,&tail->u.var.p_lpath,result);
	  break;
	case pk_neg:
	    effect_path_set_insert(effect_region,&head->u.var.n_upath,result);
	  if (tail->kind != eff_inter)
	    effect_path_set_insert(effect_region,&tail->u.var.n_lpath,result);
	  break;
	case pk_match:
	case pk_seed:
	  effect_path_set_insert(effect_region,&head->u.var.m_upath,result);
	  if (tail->kind != eff_inter)
	    effect_path_set_insert(effect_region,&tail->u.var.m_lpath,result);
	  break;
	default:
	  fail("Unexpected path kind %x\n",k);
	}


    }

  deleteregion(scratch_region);
  return TRUE;
}

static bool mkback_path(path_kind k, effect head, effect tail,
			bool reached_global)
{
  return internal_mkeffect_path(k,head,tail,reached_global,1);
}

static bool mkforward_path(path_kind k, effect head, effect tail, 
			   bool reached_global)
{
  return internal_mkeffect_path(k,head,tail,reached_global,0);
}

void forward_rule1(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_rule1: ");
  print_effect_path(printf,p);
  printf("\n");
#endif
  
  assert(tail->kind !=eff_union);
  
  if (tail->kind == eff_inter)
    return;

  assert (tail->kind == eff_constr ||
	  tail->kind == eff_var);

  scan_effect_edge_set(next_edge,eess,tail->u.var.m_ubounds)
    {
      assert(head->kind == eff_var ||
	     head->kind == eff_constr);
      
      mkforward_path(pk_match,head,next_edge->bound,
		    p->reached_global || global_effect(tail));
    }
}

void forward_rule2(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);
  
#ifdef DEBUG
  printf("f_rule2: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind == eff_inter)
    return;
  
  scan_effect_edge_set(next_edge,eess,tail->u.var.n_ubounds)
    {
      mkforward_path(pk_neg,head,next_edge->bound,
		    p->reached_global || global_effect(tail));
    }
}

void forward_rule3(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);
  
#ifdef DEBUG
  printf("f_rule3: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind == eff_inter)
    return;
  
  scan_effect_edge_set(next_edge,eess,tail->u.var.p_ubounds)
    {
      mkforward_path(pk_pos,head,next_edge->bound,
		    p->reached_global || global_effect(tail));
    }
}

void forward_rule3m(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_rule3m: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  
  if (tail->kind == eff_inter)
    return;
  
  scan_effect_edge_set(next_edge,eess,tail->u.var.p_ubounds)
    {
      mkforward_path(pk_match,head,next_edge->bound,
		    p->reached_global || global_effect(tail));
    }
}

void forward_rule4(effect_path p)
{ 
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_rule4: ");
  print_effect_path(printf,p);
  printf("\n");
#endif
  
  if (tail->kind == eff_inter)
    return;
  
  scan_effect_edge_set(next_edge,eess,tail->u.var.m_ubounds)
    {
      mkforward_path(pk_neg,head,next_edge->bound,
		    p->reached_global || global_effect(tail));
    }
}

void forward_rule5(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_rule5: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind == eff_inter)
    return;
  
  scan_effect_edge_set(next_edge,eess,tail->u.var.m_ubounds)
    {
      mkforward_path(pk_pos,head,next_edge->bound,
		    p->reached_global || global_effect(tail));
    }
}

void forward_rule6(effect_path p)
{
  
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  effect tail = ecr_effect(p->tail);
  
#ifdef DEBUG
  printf("f_rule6: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind == eff_inter)
    return;
  
  scan_effect_edge_set(next_edge,eess,tail->u.var.n_ubounds)
    {
      if (!seeded_effect(next_edge->bound))
      {
	set_seeded_effect(next_edge->bound);
	mkforward_path(pk_seed,next_edge->bound,next_edge->bound,
		      global_effect(next_edge->bound));
      }
    }
}

void new_summary(effect e1, effect e2)
{  
#ifdef DEBUG
  printf("new summary: ");
  print_effect(printf,e1);
  printf(", ");
  print_effect(printf,e2);
  printf("\n");
#endif

  internal_mkleq_effect(e1,e2,p_sub,0);
}

void forward_rule7(effect_path p)
{

  effect_edge_set_scanner lbscan,ubscan;
  effect_path_set_scanner path_scan;
  effect_edge next_lb, next_ub;
  effect_path next_path;

  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_rule7: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind == eff_inter)
    return;

  if (!param_effect(head) || !return_effect(tail))
    return;

  scan_effect_edge_set(next_lb,lbscan,head->u.var.n_lbounds)
    {
      scan_effect_edge_set(next_ub,ubscan,tail->u.var.p_ubounds)
	{
	  if (next_lb->index == next_ub->index)
	    {
	      new_summary(next_lb->bound,next_ub->bound);
	      
	       /*  rule 1  */
	      scan_effect_path_set(next_path,path_scan,head->u.var.m_lpath)
		mkforward_path(pk_match,next_path->head,next_ub->bound,
			      next_path->reached_global);
	
	      
	       /*  rule 4 */
	      scan_effect_path_set(next_path,path_scan,head->u.var.n_lpath)
		mkforward_path(pk_neg,next_path->head,next_ub->bound,
			      next_path->reached_global);
	      

	       /*  rule 5 */
	      scan_effect_path_set(next_path,path_scan,head->u.var.p_lpath)
		mkforward_path(pk_pos,next_path->head,next_ub->bound,
			      next_path->reached_global);
	      
	    }
	  
	}
    }
}

 /*  Intersections M */
void forward_inter_mrule(effect_path p)
{
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_imrule: ");
  print_effect_path(printf,p);
  printf("\n");
#endif


  if (tail->kind != eff_inter)
    return;

  assert(tail->kind == eff_inter);

 
   /*  reached slot 1 */
  if (tail == tail->u.inter_->slot_1)
    {
       /* insert into reached_1, check reached_2 */
      peffect_set_insert(effect_region,&tail->u.inter_->reached_1m,head);

      if (peffect_set_member(tail->u.inter_->reached_2m,head))
	{
#ifdef DEBUG
	  printf("Through intersection: ");
#endif
	  mkforward_path(pk_match,head,tail->u.inter_->succ,
			p->reached_global);
	}
    }
   /*  reached slot 2 */
  else if (tail == tail->u.inter_->slot_2)
    {
       /* insert into recahed_2, check reached_1 */
      peffect_set_insert(effect_region,&tail->u.inter_->reached_2m,head);
      
      if (peffect_set_member(tail->u.inter_->reached_1m,head))
	{
#ifdef DEBUG
	  printf("Through intersection: ");
#endif
	  mkforward_path(pk_match,head,tail->u.inter_->succ,
			p->reached_global);
	}
    }
  else
    {
      assert(0);
    }
}


void forward_inter_nrule(effect_path p)
{
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_inrule: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind != eff_inter)
    return;

  assert(tail->kind == eff_inter);

 
   /*  reached slot 1 */
  if (tail == tail->u.inter_->slot_1)
    {
       /* insert into reached_1, check reached_2 */
      peffect_set_insert(effect_region,&tail->u.inter_->reached_1n,head);

      if (peffect_set_member(tail->u.inter_->reached_2n,head))
	mkforward_path(pk_neg,head,tail->u.inter_->succ,
		      p->reached_global);
    }
   /*  reached slot 2 */
  else if (tail == tail->u.inter_->slot_2)
    {
       /* insert into reached_2, check reached_1 */
      peffect_set_insert(effect_region,&tail->u.inter_->reached_2n,head);
      
      if (peffect_set_member(tail->u.inter_->reached_1n,head))
	mkforward_path(pk_neg,head,tail->u.inter_->succ,
		      p->reached_global);
    }
  else
    assert(0);
}

void forward_inter_prule(effect_path p)
{

  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("f_iprule: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  if (tail->kind != eff_inter)
    return;

  assert(tail->kind == eff_inter);

 
   /*  reached slot 1 */
  if (tail == tail->u.inter_->slot_1)
    {
       /* insert into reached_1, check reached_2 */
      peffect_set_insert(effect_region,&tail->u.inter_->reached_1p,head);

      if (peffect_set_member(tail->u.inter_->reached_2p,head))
	mkforward_path(pk_pos,head,tail->u.inter_->succ,
		      p->reached_global);
    }
   /*  reached slot 2 */
  else if (tail == tail->u.inter_->slot_2)
    {
       /* insert into reached_2, check reached_1 */
      peffect_set_insert(effect_region,&tail->u.inter_->reached_2m,head);
      
      if (peffect_set_member(tail->u.inter_->reached_1p,head))
	mkforward_path(pk_pos,head,tail->u.inter_->succ,
		      p->reached_global);
    }
  else
    assert(0);
}

void matched_forward_rules(effect_path p)
{

   forward_rule1(p);
  forward_rule2(p);

if (!p->reached_global)
    forward_rule3(p);
  else
    forward_rule3m(p);
  forward_rule6(p);
  forward_rule7(p);
  forward_inter_mrule(p);
}

void negative_forward_rules(effect_path p)
{
  forward_rule2(p);
  forward_rule4(p);
  forward_rule6(p);
  forward_rule7(p);
  forward_inter_nrule(p);
}

void positive_forward_rules(effect_path p)
{
  forward_rule2(p);
  forward_rule3(p);
  forward_rule5(p);
  forward_rule6(p);
  forward_rule7(p);
  forward_inter_prule(p);
}

static void forward_reachability(void)
{
  while (!dd_is_empty(forward_path_worklist))
    {
      dd_list_pos first = dd_first(forward_path_worklist);
      effect_path p = DD_GET(effect_path,first);
      dd_remove(first);

      switch (p->kind)
	{
	case pk_pos:
	  positive_forward_rules(p);
	  break;
	case pk_neg:
	  if (global_effect(p->tail))
	      matched_forward_rules(p);
	  else
	    negative_forward_rules(p);
	  break;
	default:
	  matched_forward_rules(p);
	}
    }
}

void forward_tabulate_effects(effect e)
{
  e = ecr_effect(e);

  if (e == NULL)
    return;
  
  assert (e->kind != eff_inter);

#ifdef DEBUG
  printf("Forward tabulating effects from: ");
  print_effect(printf,e);
  printf("\n");
#endif
  
  mkforward_path(pk_seed,e,e,global_effect(e));

  forward_reachability();
}

void backwards_rule1(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;
  effect tail = p->tail;
  
#ifdef DEBUG
  printf("b_rule1: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  assert(head->kind == eff_constr);
  
  
  scan_effect_edge_set(next_edge,eess,head->u.var.m_lbounds)
    {
      mkback_path(pk_match,next_edge->bound,tail, 
		   p->reached_global || global_effect(head));
    }
  
}

void backwards_rule2(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;
  effect tail = p->tail;
  
#ifdef DEBUG
  printf("b_rule2: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  assert(head->kind == eff_constr);
  
  
  scan_effect_edge_set(next_edge,eess,head->u.var.n_lbounds)
    {
      mkback_path(pk_neg,next_edge->bound,tail, 
		   p->reached_global || global_effect(head));
    }
  
}

void backwards_rule2m(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;
  effect tail = p->tail;

#ifdef DEBUG
  printf("b_rule2m: ");
  print_effect_path(printf,p);
  printf("\n");
#endif
  
  assert(head->kind == eff_constr);
  
  
  scan_effect_edge_set(next_edge,eess,head->u.var.n_lbounds)
    {
      mkback_path(pk_match,next_edge->bound,tail, 
		   p->reached_global || global_effect(head));
    }
}

void backwards_rule3(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;
  effect tail = p->tail;
  
#ifdef DEBUG
  printf("b_rule3: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  assert(head->kind == eff_constr);
  
  
  scan_effect_edge_set(next_edge,eess,head->u.var.p_lbounds)
    {
      mkback_path(pk_pos,next_edge->bound,tail, 
		   p->reached_global || global_effect(head));
    }
}

void backwards_rule4(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;
  effect tail = p->tail;
  
#ifdef DEBUG
  printf("b_rule4: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  assert(head->kind == eff_constr);
  
  
  scan_effect_edge_set(next_edge,eess,head->u.var.m_lbounds)
    {
      mkback_path(pk_neg,next_edge->bound,tail, 
		   p->reached_global || global_effect(head));
    }

}

void backwards_rule5(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;
  effect tail = p->tail;
  
#ifdef DEBUG
  printf("b_rule5: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  assert(head->kind == eff_constr);
  
  scan_effect_edge_set(next_edge,eess,head->u.var.m_lbounds)
    {
      mkback_path(pk_pos,next_edge->bound,tail, 
		   p->reached_global || global_effect(head));
    }

}

void backwards_rule6(effect_path p)
{
  effect_edge_set_scanner eess;
  effect_edge next_edge;
  
  effect head = p->head;

#ifdef DEBUG
  printf("b_rule6: ");
  print_effect_path(printf,p);
  printf("\n");
#endif
  
  assert(head->kind == eff_constr);

  scan_effect_edge_set(next_edge,eess,head->u.var.p_lbounds)
    {
      if (!seeded_effect(next_edge->bound))
      {
	set_seeded_effect(next_edge->bound);
	mkforward_path(pk_seed,next_edge->bound,next_edge->bound,
		      global_effect(next_edge->bound));
      }
    }
}

void backwards_rule7(effect_path p)
{
  effect_edge_set_scanner lbscan,ubscan;
  effect_path_set_scanner path_scan;
  effect_edge next_lb, next_ub;
  effect_path next_path;
  
  effect head = ecr_effect(p->head);
  effect tail = ecr_effect(p->tail);

#ifdef DEBUG
  printf("b_rule7: ");
  print_effect_path(printf,p);
  printf("\n");
#endif

  assert(head->kind == eff_constr);
  
  if (!param_effect(head) || !return_effect(tail))
    return;

  scan_effect_edge_set(next_lb,lbscan,head->u.var.n_lbounds)
    {
      scan_effect_edge_set(next_ub,ubscan,tail->u.var.p_ubounds)
	{
	  if (next_lb->index == next_ub->index)
	    {
	      new_summary(next_lb->bound,next_ub->bound);
	      
	       /*  rule 1  */
	      scan_effect_path_set(next_path,path_scan,tail->u.var.m_upath)
		mkforward_path(pk_match,next_path->tail,next_lb->bound,
			      next_path->reached_global);

	       /*  rule 4 */
	      scan_effect_path_set(next_path,path_scan,tail->u.var.n_upath)
		mkforward_path(pk_neg,next_path->tail,next_lb->bound,
			      next_path->reached_global);

	       /*  rule 5 */
	      scan_effect_path_set(next_path,path_scan,tail->u.var.p_upath)
		mkforward_path(pk_pos,next_path->tail,next_lb->bound,
			      next_path->reached_global);
	    }
	  
	}
    }

}

void matched_backward_rules(effect_path p)
{
  backwards_rule1(p);
  if (!p->reached_global)
    backwards_rule2(p);
  else
    backwards_rule2m(p);
  backwards_rule3(p);
  backwards_rule6(p);
  backwards_rule7(p);
  
}

void negative_backward_rules(effect_path p)
{
  backwards_rule2(p);
  backwards_rule3(p);
  backwards_rule4(p);
  backwards_rule6(p);
  backwards_rule7(p);
}

void positive_backward_rules(effect_path p)
{
  backwards_rule3(p);
  backwards_rule5(p);
  backwards_rule6(p);
  backwards_rule7(p);
}

void backward_reachability(void)
{
  while (! dd_is_empty(back_path_worklist))
    {
      dd_list_pos first = dd_first(back_path_worklist);
      effect_path p = DD_GET(effect_path,first);
      dd_remove(first);

      switch (p->kind)
	{
	case pk_pos:
	  if (global_effect(p->tail))
	    matched_backward_rules(p);
	  else
	    positive_backward_rules(p);
	  break;
	case pk_neg:
	  negative_backward_rules(p);
	  break;
	default:
	  matched_backward_rules(p);
	}
    }

}


 /*   compute all aliases of e (e is a constructed effect, i.e. an aloc) */
void backward_tabulate_aliases(effect e)
{
  aloc_set_scanner scan;
  aloc next_aloc,al;
  e = ecr_effect(e);

  if (e == NULL)
    return;
  assert(e->kind == eff_constr);

#ifdef DEBUG
  printf("Backwards tabulating aliases for: ");
  print_effect(printf,e);
  printf("\n");
#endif

  mkback_path(pk_seed,e,e,global_effect(e));
  

  al = ecr_aloc(e->u.constr.base);

   /*  then for each alias, compute forward reachability of constructed effects */
  scan_aloc_set(next_aloc,scan,al->u.var.alias_cache)
    {
      int i;
      next_aloc = ecr_aloc(next_aloc);
      assert(next_aloc->u.var.effects);
      
      for (i = 0; i <= eff_last; i++)
	{
	  if (next_aloc->u.var.effects[i])
	    forward_tabulate_effects(next_aloc->u.var.effects[i]);
	}
    }
}

static unsigned int compute_mark_through_aliases(aloc al, effect e)
{
  unsigned int mark =0 ,temp_mark = 0;
  aloc_set_scanner scan;
  aloc next_aloc;
  
  al = ecr_aloc(al);
  e = ecr_effect(e);

  assert(e->kind != eff_union);
  assert(e->kind != eff_inter);

  /* check only if something reached e */
  if (e->u.var.reach_cache)
    {
#ifdef DEBUG
      print_aloc(printf, al);
      printf(" -?-> ");
      print_effect(printf, e);
      printf("\n");
#endif


      /* al itself reaches e */
      aloc_int_map_lookup(e->u.var.reach_cache, al, &mark);
      
      /* or aliases of al reach e */
      scan_aloc_set(next_aloc,scan,al->u.var.alias_cache)
	{
	  aloc_int_map_lookup(e->u.var.reach_cache,next_aloc,&temp_mark);
	  mark |= temp_mark;
	}
    }
  return mark;
}

/* al is in goal if al or any alias of al reaches goal */
static int aloc_in_effect(aloc al, effect goal)
{
  unsigned int mark;
  int i;
  effect *effects;

  assert(state == state_reach);
  
  al = ecr_aloc(al);
  assert(al->kind == aloc_var);

  goal = ecr_effect(goal);
  assert(goal->kind != eff_inter);

  effects = al->u.var.effects;
  mark = 0;

  if (! restrict_qual && ! flag_flow_sensitive)
    return FALSE;
  
  if (!effects)
    return FALSE;
  else
    {
      for (i = 0; i <= eff_last; i++)
	{
	  if (effects[i])
	    {
	      forward_tabulate_effects(effects[i]);
	      backward_tabulate_aliases(effects[i]);
	    }
	}
      mark = compute_mark_through_aliases(al,goal);
    }
  
  return mark;
}

#ifdef DEBUG
static char *eff_kind_to_str(eff_kind k)
{     
  switch (k)
    {
    case eff_any:
      return "any";
      break;
    case eff_rwr:
      return "r/wr";
      break;
    case eff_r:
      return "read";
      break;
    case eff_wr:
      return "write";
      break;
    case eff_alloc:
      return "alloc";
      break;
    default:
      fail("Unexpected constraint kind %x\n", k);
    }
}
#endif

/* The constraint e1 <= e2 (subtyping) was added during reachability. 
   Extend effect and alias paths involving e1 and e2 by scanning through 
   all hashed paths */

/*
void extend_effect_paths(effect e1, effect e2)
{
  sig_effect_path_map_scanner scan;
  int *key;
  effect_path p;

  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);

  
  assert(state == state_reach);
  assert(e1->kind == eff_constr || e1->kind == eff_var);
  assert(e2->kind == eff_constr || e2->kind == eff_var);

  sig_effect_path_map_scan(back_paths,&scan);
  
  while(sig_effect_path_map_next(&scan,&key,&p))
    {
      if (eq_effect(e1,p->tail))
	mkback_path(p->kind,p->head,e2,p->reached_global);
      if (eq_effect(e2,p->head))
	mkback_path(p->kind,e1,p->tail,p->reached_global);
    }

  sig_effect_path_map_scan(forward_paths,&scan);

  while(sig_effect_path_map_next(&scan,&key,&p))
    {
      if (eq_effect(e1,p->tail))
	mkforward_path(p->kind,p->head,e2,p->reached_global);
      if (eq_effect(e2,p->head))
	mkforward_path(p->kind,e1,p->tail,p->reached_global);
    }

  backward_reachability();
  forward_reachability();
}
*/

/* The constraint e1 <= e2 (subtyping) was added during reachability. 
   Extend effect and alias paths involving e1 and e2 by scanning through 
   all hashed paths */
void extend_effect_paths(effect e1, effect e2)
{
  effect_path_set_scanner path_scan;
  effect_path p;

  e1 = ecr_effect(e1);
  e2 = ecr_effect(e2);

  assert(state == state_reach);
  assert(e1->kind == eff_constr || e1->kind == eff_var);
  assert(e2->kind == eff_constr || e2->kind == eff_var);
  
  scan_effect_path_set(p,path_scan,e2->u.var.p_upath)  
    {
      assert(eq_effect(p->head,e2));
      mkback_path(p->kind,e1,p->tail,p->reached_global);
    }
  scan_effect_path_set(p,path_scan,e2->u.var.m_upath)
    {
      assert(eq_effect(p->head,e2));
      mkback_path(p->kind,e1,p->tail,p->reached_global);
    }

  scan_effect_path_set(p,path_scan,e2->u.var.n_upath)
    {
      assert(eq_effect(p->head,e2));
      mkback_path(p->kind,e1,p->tail,p->reached_global);
    }

  scan_effect_path_set(p,path_scan,e1->u.var.n_lpath)
    {
      assert(eq_effect(p->tail,e1));
      mkforward_path(p->kind,p->head,e2,p->reached_global);
    }

 scan_effect_path_set(p,path_scan,e1->u.var.p_lpath)
    {
      assert(eq_effect(p->tail,e1));
      mkforward_path(p->kind,p->head,e2,p->reached_global);
    }

 scan_effect_path_set(p,path_scan,e1->u.var.m_lpath)
    {
      assert(eq_effect(p->tail,e1));
      mkforward_path(p->kind,p->head,e2,p->reached_global);
    }
 
 backward_reachability();
 forward_reachability();
}

/* TODO -- check confine constraints */
void check_nin(void)
{
  dd_list_pos cur;
  
  assert(state == state_reach);



  /* Forall constraints */
  dd_scan(cur,forall_constraints)
    {
      forall_nin_constraint c = DD_GET(forall_nin_constraint, cur);
      if (!c->failed)
	{
	  effect e = ecr_effect(c->eff);
	  bool error = FALSE;

	  assert(e->kind == eff_var);
	  
	  mkback_path(pk_seed,e,e,FALSE);
	  backward_reachability();
	  
	  /* First, check that reachability succeeded */
	  if (e->u.var.reach_failed)
	    error = TRUE;
	  /* Now check that none of the effect kind reached e's cache */
	  else if (e->u.var.reach_cache)
	    {
	      unsigned int mark;
	      aloc_int_map_scanner scan;
	      aloc al;

	      aloc_int_map_scan(e->u.var.reach_cache,&scan);

	      while (aloc_int_map_next(&scan,&al,&mark))
		{
		  if (mark & eff_kind_to_mark(c->kind))
		    {
		      error = TRUE;
		      break;
		    }
		}
	    }


	  if (error && !c->inference)
	      report_qerror(c->loc, sev_err, 
			    "Invalid use of restrict - side effect");
	}
    }

  /* Forall cond constraints */
  dd_scan(cur,forall_cond_constraints)
    {
      forall_cond_nin_constraint c = DD_GET(forall_cond_nin_constraint,cur);
      
      if (!c->failed)
	{
	  effect e1 = ecr_effect(c->eff1);
	  effect e2 = ecr_effect(c->eff2);
	  bool error = FALSE;

	  assert(e1->kind == eff_var);
	  assert(e2->kind == eff_var);

	  mkback_path(pk_seed,e1,e1,FALSE);
	  backward_reachability();

	  /* First, check that reachability succeeded */
	  if (e1->u.var.reach_failed)
	    error = TRUE;
	  /* Now for k(rho) that reached e1's cache, make
	     sure that k(rho) is not in e2 with a new 
	     disinclusion constraint */
	  else if (e1->u.var.reach_cache)
	    {
	      unsigned int mark;
	      aloc_int_map_scanner scan;
	      aloc al;

	      aloc_int_map_scan(e1->u.var.reach_cache,&scan);

	      while (aloc_int_map_next(&scan,&al,&mark))
		{
		  if (mark & (eff_kind_to_mark(c->kind1)))
		    {
		      nin_constraint nc = 
			internal_mknin_aloc_effect(c->loc,al,c->kind2,e2);
		      nc->from_forall = TRUE;
		    }
		}
	    }

	  if (error && !c->inference)
	    report_qerror(c->loc,sev_err,
			  "Invalid use of restrict - readonly written");
	  
	}

    }

  /* Disinclusion constraints */
  dd_scan(cur,nin_constraints)
    {
      nin_constraint c;
      c = DD_GET(nin_constraint,cur);
      if (!c->failed)
	{
	  unsigned int mask;
	  
	  mask = eff_kind_to_mark(c->k);
	  
#ifdef DEBUG
	  print_aloc(printf, c->aloc);
	  printf(" ! -%s-> ",eff_kind_to_str(c->k));
	  print_effect(printf, c->eff);
	  printf("\n");
#endif
	  if (aloc_in_effect(c->aloc,c->eff) & mask)
	    {
#ifdef DEBUG
	      printf("Failed constraint : ");
	      print_aloc(printf, c->aloc);
	      printf(" ! -%s-> ",eff_kind_to_str(c->k));
	      print_effect(printf, c->eff);
	      printf("\n");
#endif
	      if (!c->inference)
		{
		  if (c->from_forall)
		    report_qerror(c->loc,sev_err,
				  "Invalid use of restrict - readonly written");
		  else
		    report_qerror(c->loc, sev_err,
				  "Invalid use of restrict");
		}
	      c->failed = TRUE;
	    }
	}
    }
}

int print_effect(printf_func pf, effect e)
{
 int result;

  result = 0;
  e = ecr_effect(e);
  if (e == NULL)
    return pf("0");
  switch(e->kind)
    {
    case eff_constr:
      switch (e->u.constr.kind)
	{
	case eff_any:
	  break;
	case eff_alloc:
	  result += pf("alloc(");
	  break;
	case eff_rwr:
	  result += pf("rwr(");
	  break;
	case eff_r:
	  result += pf("r(");
	  break;
	case eff_wr:
	  result += pf("wr(");
	  break;
	default:
	  fail("Unexpected constructor kind %x\n", e->u.constr.kind);
	}
      result += print_unique_aloc(pf, e->u.constr.base);
      switch (e->u.constr.kind)
	{
	case eff_alloc:
	case eff_rwr:
	  result += pf(")");
	  break;
	case eff_r:
	  result += pf(")");
	  break;
	case eff_wr:
	  result += pf(")");
	  break;
	case eff_any:
	  break;
	default:
	  fail("Unexpected constructor kind %x\n", e->u.constr.kind);
	}
      break;
    case eff_union:
      result += print_effect(pf, e->u.union_.e1);
      result += pf(" + ");
      result += print_effect(pf, e->u.union_.e2);
      break;
    case eff_inter:
      result += pf("[(");
      result += print_effect(pf, e->u.inter_->e1);
      result += pf(") & (");
      result += print_effect(pf, e->u.inter_->e2);
      result += pf(")]");
      break;
    case eff_var:
      if (flag_ugly)
	result += pf("%s %p", e->u.var.name, e);
      else
	result += pf("%s", e->u.var.name);
      break;
    default:
      fail("Unexpected effect kind %x\n", e->kind);
    }
  return result;
  
}

#ifdef DEBUG
static char *polarity_to_string(polarity p)
{
  char *result;
  switch (p)
    {
    case p_sub: 
      result = "M";
      break;
    case p_neg: 
      result = "N";
      break;
    case p_pos: 
      result = "P";
      break;
    case p_non: 
      result = "T";
      break;
    default:
      fail("Unexpected polarity %x\n", p);
    }
  
  return result;
}
#endif


#ifdef DEBUG
static char *path_kind_to_string(path_kind k)
{
  char *result;
  
  switch (k)
    {
    case pk_seed: 
      result = "S";
      break;
    case pk_match: 
      result = "M";
      break;
    case pk_pos: 
      result = "P";
      break;
    case pk_neg: 
      result = "N";
      break;
    default:
      fail("Unexpected path kind %x\n",k);
    }
  
  return result;
}
#endif

/*
static int print_effect_edge(printf_func pf, effect_edge e, polarity p)
{
  int result;
  
  result = 0;
 
  result += pf(" -- %s(%d) --> ",polarity_to_string(p),e->index);
  result += print_effect(pf,e->bound);
  
  return result;
}
*/


#ifdef DEBUG
static int print_effect_path(printf_func pf, effect_path p)
{
  int result;

  result = 0;

  result += print_effect(pf,p->head);
  result += pf(" -- %s --> ", path_kind_to_string(p->kind));
  result += print_effect(pf,p->tail);
  if (p->reached_global)
    result += pf(" (rg) ");

  return result;
}
#endif

void print_effect_graph(void)
{
}

/**************************************************************************
 *                                                                        *
 * Stores                                                                 *
 *                                                                        *
 **************************************************************************/

bool eq_store(store s1, store s2)
{
  return FALSE;
}

void init_store(void)
{
}

store make_store_var(location loc, const char *name)
{
  return NULL;
}

store make_store_filter(location loc, store s, effect e)
{
  return NULL;
}

store make_store_ref(location loc, store s, aloc al)
{
  return NULL;
}

store make_store_ref_effect(location loc, store s, effect e)
{
  return NULL;
}

store make_store_ow(location loc, store s1, store s2, effect e)
{
  return NULL;
}

store make_store_assign(location loc, const char *err_msg, store s, aloc al,
			qtype qt, bool strong)
{
  return NULL;
}

void mkleq_store(location loc, const char *err_msg, store left, store right)
{
}

void mkeq_store(location loc, const char *err_msg, store s1, store s2)
{
}

store lub_store(location loc, const char *err_msg, store s1, store s2)
{
  return NULL;
}

void unify_store(location loc, const char *err_msg, store s1, store s2)
{
}

void propagate_store_cell_forward(store s, aloc al)
{
}

void propagate_store_cell_backward(store s, aloc al)
{
}

void compute_lins(void)
{
}

int lin_from_store(store s, aloc al)
{
  return 0;
}

qtype qtype_from_store(location loc, store s, aloc al)
{
  return NULL;
}

int print_store(printf_func pf, store s)
{
  return 0;
}
int print_store_cells(printf_func pf, pr_qual_fn pr_qual, store s,
                      bool print_alocs)
{
  return 0;
}
int print_store_raw(printf_func pf, store s)
{
  return 0;
}

