/* This file is part of cqual.
   Copyright (C) 2005 The University of Maryland.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. All advertising materials mentioning features or use of this software
     must display the following acknowledgement:
 	This product includes software developed by the University of
 	California, Berkeley and its contributors.
  4. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. */

#ifdef HAVE_LIBEXPAT

#include <expat.h>
#include <stdio.h>
#include <string.h>
#include "xmldb.h"
#include "dd_list.h"
#include "parser.h"

#define DB_BUFSIZE 8192

/*********************************************************************/

/* Parser for an app */

typedef struct Bin_state {
  FILE *db;			/* datafile file handle */
  XML_Parser parser;		/* XML parser */
  bool at_end;			/* True if db has been fully parsed */
  char db_buf[DB_BUFSIZE];	/* Parser buffer */
  char *bin_name;		/* Name of application */
  int in_bin;			/* True if we are reading data for bin_name */
  dd_list queue;		/* Source file queue */
} *bin_state;

static void bin_start_executable(bin_state state, const char **attr) {
  int i;

  for (i = 0; attr[i]; i += 2)
    if (!strcmp(attr[i], "binName") &&
	!strcmp(attr[i+1], state->bin_name))
      state->in_bin = 1;
}

static void bin_end_executable(bin_state state) {
  if (state->in_bin)
    state->in_bin = 0;
}

static void bin_start_source(bin_state state, const char **attr) {
  int i;

  if (!state->in_bin)
    return;

  for (i = 0; attr[i]; i += 2)
    if (!strcmp(attr[i], "iFile"))
      dd_add_last(parse_region, state->queue,
		  (void *) rstrdup(parse_region, attr[i+1]));
}

static void bin_start(void *data, const char *el, const char **attr) {
  bin_state state = (bin_state) data;

  if (!strcmp(el, "executable") || !strcmp(el, "sharedLib"))
    bin_start_executable(state, attr);
  else if (!strcmp(el, "source"))
    bin_start_source(state, attr);
}

static void bin_end(void *data, const char *el) {
  bin_state state = (bin_state) data;

  if (!strcmp(el, "executable") || !strcmp(el, "sharedLib"))
    bin_end_executable(state);
}

void parse_db(bin_state state) {
  enum XML_Status status;
  int eof;
  size_t len;

  /* Should never be called after everything's been parsed */
  if (state->at_end)
    return;

  len = fread(state->db_buf, 1, DB_BUFSIZE, state->db);
  if (len < 0) {
    perror("Error reading database file");
    state->at_end = TRUE;
    return;
  }
  eof = feof(state->db);
  status = XML_Parse(state->parser, state->db_buf, len, eof);
  if (status == XML_STATUS_ERROR) {
    fprintf(stderr, "XML Error:%d:%s\n",
	    XML_GetCurrentLineNumber(state->parser),
	    XML_ErrorString(XML_GetErrorCode(state->parser)));
    state->at_end = TRUE;
  }

  state->at_end = eof;
}

static bool bin_has_next(xmldb_it i) {
  bin_state state = (bin_state) i->state;

  if (!dd_is_empty(state->queue))
    return TRUE;

  while (dd_is_empty(state->queue) && !state->at_end) {
    parse_db(state);
  }

  return !dd_is_empty(state->queue);
}

static void *bin_next(xmldb_it i) {
  bin_state state = (bin_state) i->state;
  void *source_file;

  if (!bin_has_next(i))
    return NULL;
  /* Queue is now non-empty */

  source_file = DD_GET(void *, dd_first((state->queue)));
  dd_remove(dd_first((state->queue)));
  return source_file;
}

static void bin_reset(xmldb_it i) {
  bin_state state = (bin_state) i->state;

  XML_ParserReset(state->parser, NULL);
  rewind(state->db);
  state->at_end = FALSE;
  state->in_bin = FALSE;
  state->queue = dd_new_list(parse_region);
  XML_SetUserData(state->parser, state);
  XML_SetStartElementHandler(state->parser, bin_start);
  XML_SetEndElementHandler(state->parser, bin_end);
}

static void bin_destroy(xmldb_it i) {
  bin_state state = (bin_state) i->state;

  XML_ParserFree(state->parser);
  i->state = NULL;
}

/* Return an iterator that returns all source files for an
   application.  Returns NULL if no such app in database. */
xmldb_it xmldb_init_bin(char *xmldb_file, char *app) {
  bin_state state;
  xmldb_it i;

  state = (bin_state) malloc(sizeof(struct Bin_state));
  if (!(state->db = fopen(xmldb_file, "rt"))) {
    perror("Can't open database file");
    free(state);
    return NULL;
  }
  if (!(state->parser = XML_ParserCreate(NULL))) {
    fprintf(stderr, "Couldn't create XML Parser\n");
    free(state);
    return NULL;
  }
  state->at_end = FALSE;
  state->bin_name = rstrdup(parse_region, app);
  state->in_bin = FALSE;
  state->queue = dd_new_list(parse_region);
  XML_SetUserData(state->parser, state);
  XML_SetStartElementHandler(state->parser, bin_start);
  XML_SetEndElementHandler(state->parser, bin_end);


  i = (xmldb_it) malloc(sizeof(struct Xmldb_it));
  i->has_next = bin_has_next;
  i->next = bin_next;
  i->reset = bin_reset;
  i->destroy = bin_destroy;
  i->state = state;
  return i;
}

#endif /* HAVE_LIBEXPAT */
